import os
import shutil
import yaml
import pytest

from res.enkf import EnKFMain, ResConfig

import semeio
import semeio.workflows.misfit_preprocessor.misfit_preprocessor as misfit_preprocessor
from semeio.workflows.correlated_observations_scaling.exceptions import (
    EmptyDatasetException,
)
from tests.jobs.correlated_observations_scaling.conftest import TEST_DATA_DIR

from unittest.mock import Mock


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_misfit_preprocessor_main_entry_point_gen_data(monkeypatch):
    run_mock = Mock()
    scal_job = Mock(return_value=Mock(run=run_mock))
    monkeypatch.setattr(
        misfit_preprocessor,
        "CorrelatedObservationsScalingJob",
        scal_job,
    )

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    config = {"clustering": {"spearman_correlation": {"fcluster": {"t": 1.0}}}}
    config_file = "my_config_file.yaml"
    with open(config_file, "w") as f:
        yaml.dump(config, f)

    misfit_preprocessor.MisfitPreprocessorJob(ert).run(config_file)

    # call_args represents the clusters, we expect the snake_oil
    # observations to generate this amount of them
    # call_args is a call object, which itself is a tuple of args and kwargs.
    # In this case, we want args, and the first element of the arguments, which
    # again is a tuple containing the configuration which is a list of configs.
    assert len(list(run_mock.call_args)[0][0]) == 47, "wrong number of clusters"


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_misfit_preprocessor_passing_scaling_parameters(monkeypatch):
    run_mock = Mock()
    scal_job = Mock(return_value=Mock(run=run_mock))
    monkeypatch.setattr(
        misfit_preprocessor,
        "CorrelatedObservationsScalingJob",
        scal_job,
    )

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    config = {"scaling": {"threshold": 0.5, "std_cutoff": 2, "alpha": 3}}
    config_file = "my_config_file.yaml"
    with open(config_file, "w") as f:
        yaml.dump(config, f)

    misfit_preprocessor.MisfitPreprocessorJob(ert).run(config_file)

    for scaling_config in list(run_mock.call_args)[0][0]:
        assert 0.5 == scaling_config["CALCULATE_KEYS"]["threshold"]


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_misfit_preprocessor_main_entry_point_no_config(monkeypatch):
    run_mock = Mock()
    scal_job = Mock(return_value=Mock(run=run_mock))
    monkeypatch.setattr(
        misfit_preprocessor,
        "CorrelatedObservationsScalingJob",
        scal_job,
    )

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    misfit_preprocessor.MisfitPreprocessorJob(ert).run()

    assert len(run_mock.call_args[0][0]) > 1  # pylint: disable=unsubscriptable-object


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_misfit_preprocessor_with_scaling():
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    config = {"clustering": {"spearman_correlation": {"fcluster": {"t": 1.0}}}}
    config_file = "my_config_file.yaml"
    with open(config_file, "w") as f:
        yaml.dump(config, f)

    misfit_preprocessor.MisfitPreprocessorJob(ert).run(config_file)

    # assert that this arbitrarily chosen cluster gets scaled as expected
    obs = ert.getObservations()["FOPR"]
    for index in [13, 14, 15, 16, 17, 18, 19, 20]:
        assert obs.getNode(index).getStdScaling() == 2.8284271247461903


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_misfit_preprocessor_skip_clusters_yielding_empty_data_matrixes(monkeypatch):
    def raising_scaling_job(data):
        if data == {"CALCULATE_KEYS": {"keys": [{"index": [88, 89], "key": "FOPR"}]}}:
            raise EmptyDatasetException("foo")

    scaling_mock = Mock(return_value=Mock(**{"run.side_effect": raising_scaling_job}))
    monkeypatch.setattr(
        misfit_preprocessor, "CorrelatedObservationsScalingJob", scaling_mock
    )

    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    config = {"clustering": {"spearman_correlation": {"fcluster": {"t": 1.0}}}}
    config_file = "my_config_file.yaml"
    with open(config_file, "w") as f:
        yaml.dump(config, f)

    job = misfit_preprocessor.MisfitPreprocessorJob(ert)

    try:
        job.run(config_file)
    except EmptyDatasetException:
        pytest.fail("EmptyDatasetException was not handled by misfit preprocessor")


@pytest.mark.skipif(TEST_DATA_DIR is None, reason="no libres test-data")
@pytest.mark.usefixtures("setup_tmpdir")
def test_misfit_preprocessor_invalid_config():
    test_data_dir = os.path.join(TEST_DATA_DIR, "local", "snake_oil")

    shutil.copytree(test_data_dir, "test_data")
    os.chdir(os.path.join("test_data"))

    res_config = ResConfig("snake_oil.ert")
    ert = EnKFMain(res_config)

    config = {
        "unknown_key": [],
        "clustering": {"spearman_correlation": {"fcluster": {"threshold": 1.0}}},
    }
    config_file = "my_config_file.yaml"
    with open(config_file, "w") as f:
        yaml.dump(config, f)

    job = misfit_preprocessor.MisfitPreprocessorJob(ert)
    with pytest.raises(semeio.workflows.misfit_preprocessor.ValidationError) as ve:
        job.run(config_file)

    expected_err_msg = (
        "Invalid configuration of misfit preprocessor\n"
        "  - Unknown key: unknown_key (root level)\n"
        "  - Unknown key: threshold (clustering.spearman_correlation.fcluster)\n"
    )
    assert expected_err_msg == str(ve.value)



import os

import logging
import pandas as pd

logger = logging.getLogger(__name__)


def _write_gen_data_files(trajectory_df, fname):
    """Generate three files with the information GEN_DATA needs
    from the trajectory dataframe.

    See https://github.com/equinor/ert/blob/0dc96c49ca8eafee54a06227530b5a47b094e2bc/docs/tips.txt#L87

    One main file (fname) will be produced with pressure values, and two
    auxiliary files will be produced with the suffixes "_active" and
    "_inactive_info". The _active file tells ERT if some of the values
    in the base file should be ignored, and _inactive_info will give
    an explanation for why it should be ignored.

    Args:
        trajectory_df (pd.DataFrame): The column "order" should contain
            the original row order from input text files. "pressure" should
            hold pressure data, but will be defaulted if not present. "is_active"
            should be a boolean column, and "inactive_info" should contain
            strings with information on why points are inactive.
        fname (str): Filename to hold the values for the GEN_DATA keyword,
            which are pressure values in this case, and used as a basename
            for the two auxiliary files.
    """  # noqa
    _write_pressure(fname, trajectory_df)
    _write_active(fname + "_active", trajectory_df)
    _write_inactive_info(fname + "_inactive_info", trajectory_df)


def _write_pressure(fname, trajectory_df):
    """Write pressure value, one pr line for all points, -1 is used where
    there is no pressure information.
    """
    with open(fname + "", "w+") as fh:
        if "pressure" in trajectory_df:
            fh.write(
                "\n".join(
                    trajectory_df.sort_values("order")["pressure"]
                    .fillna(value=-1)
                    .astype(str)
                    .values
                )
                + "\n"
            )
        else:
            fh.write("\n".join(["-1" * len(trajectory_df)]))

    logger.info("Forward model script gendata_rft.py: Wrote file {}".format(fname))


def _write_active(fname, trajectory_df):
    """Write a file with "1" pr row if a point is active, "0" if not"""
    with open(fname, "w+") as fh:
        fh.write(
            "\n".join(
                trajectory_df.sort_values("order")["is_active"]
                .astype(int)
                .astype(str)
                .values
            )
        )


def _write_inactive_info(fname, trajectory_df):
    """Write a file with explanations to users for inactive points"""
    with open(fname, "w+") as fh:
        if "inactive_info" not in trajectory_df:
            fh.write("")
        else:
            fh.write(
                "\n".join(
                    trajectory_df[~trajectory_df["is_active"]]
                    .sort_values("order")["inactive_info"]
                    .dropna()
                    .values
                )
            )


def _populate_trajectory_points(
    well, date, trajectory_points, ecl_grid, ecl_rft, zonemap=None
):
    """
    Populate a list of trajectory points, that only contain UTM coordinates
    for a well-path, with (i,j,k) indices corresponding to a given Eclipse grid,
    and simulated pressure at the given date in the cells.

    Args:
        well (str): Eclipse-name of well
        date (datetime.date): Date for which the RFT observation should be taken,
            must correspond to a date in the Eclipse RFT binary output
        trajectory_points (list of points): Representing the wellpath in UTM at which
            the real life RFT observations points are valid.
        ecl_grid (libecl Grid object):
        ecl_rft (libecl RFT object):
        zonemap (dict):

    Returns:
        The list of trajectory points, augmented with i,j,k, simulated pressure
        and the zone name if zonemap is provided.
    """
    try:
        rft = ecl_rft.get(well, date)
    except KeyError:
        logger.error(
            (
                "Forward model script gendata_rft.py: "
                "No RFT data written for well ({}) and date ({})"
            ).format(well, date)
        )
        return []

    for point in trajectory_points:
        point.set_ijk(
            ecl_grid.find_cell(point.utm_x, point.utm_y, point.true_vertical_depth)
        )
        point.update_pressure_from_rft(rft)
        point.validate_zone(zonemap)

    return trajectory_points


def run(
    well_times,
    trajectories,
    ecl_grid,
    ecl_rft,
    zonemap=None,
    csvfile=None,
    outputdirectory=".",
):
    dframes = []

    if not well_times:
        raise ValueError("No RFT data requested")

    for well, time, report_step in well_times:
        logger.debug(
            "Collecting RFT for well: {} at date: {}, report step: {}".format(
                well, time, report_step
            )
        )

        trajectory_points = _populate_trajectory_points(
            well, time, trajectories[well], ecl_grid, ecl_rft, zonemap
        )

        if trajectory_points:
            # Aggregate the same data to a dataframe,
            # each trajectory tagged by well and time:
            trajectory_df = trajectory_points.to_dataframe(zonemap=zonemap).assign(
                well=well, time=time
            )

            # Write trajectory and associated data to three  files,
            # individual pr. well and time, this is for GEN_DATA to pick up.
            _write_gen_data_files(
                trajectory_df,
                os.path.join(outputdirectory, "RFT_{}_{}".format(well, report_step)),
            )

            # Aggregate dataframe for all wells and report steps.
            dframes.append(trajectory_df)

    if csvfile is not None:
        pd.concat(dframes, ignore_index=True, sort=False).to_csv(csvfile, index=None)

    if len(dframes) < len(well_times):
        raise ValueError("Failed to extract requested RFT data")



from semeio.workflows.misfit_preprocessor.exceptions import ValidationError
from semeio.workflows.misfit_preprocessor.config import (
    MisfitPreprocessorConfig,
    assemble_config,
)
from semeio.workflows.misfit_preprocessor.job import run

__all__ = (
    "ValidationError",
    "MisfitPreprocessorConfig",
    "assemble_config",
    "run",
)



"""Test module for ERT-Pyscal integration"""
import os
import sys
import random

import pytest

import pandas as pd

from semeio.jobs.scripts.fm_pyscal import run, main_entry_point


EXAMPLE_STATIC_DFRAME = pd.DataFrame(
    columns=["SATNUM", "Nw", "Now", "Nog", "Ng"],
    data=[[1, 2, 2, 3, 3], [2, 4, 4, 1, 1]],
)
EXAMPLE_SCAL = pd.DataFrame(
    columns=["SATNUM", "CASE", "Nw", "Now", "Nog", "Ng"],
    data=[[1, "Low", 2, 2, 3, 3], [1, "Base", 4, 4, 1, 1], [1, "High", 5, 5, 2, 2]],
)
EXAMPLE_WATEROIL = pd.DataFrame(columns=["SATNUM", "Nw", "NOW"], data=[[1, 2, 2]])


@pytest.mark.parametrize(
    "dframe, runargs",
    [
        (EXAMPLE_STATIC_DFRAME, ["__NONE__", "__NONE__", "__NONE__", "sgof", 1]),
        (EXAMPLE_STATIC_DFRAME, ["__NONE__", "__NONE__", "__NONE__", "slgof", 1]),
        (EXAMPLE_STATIC_DFRAME, ["__NONE__", "__NONE__", "__NONE__", "slgof", 2]),
        (EXAMPLE_STATIC_DFRAME, ["__NONE__", "__NONE__", "__NONE__", "sgof", 2]),
        (EXAMPLE_WATEROIL, ["__NONE__", "__NONE__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "INTERPOLATE_WO", "INTERPOLATE_GO", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "INTERPOLATE_WO", "INTERPOLATE_GO", "slgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "INTERPOLATE_WO", "INTERPOLATE_GO", "sgof", 2]),
        (EXAMPLE_SCAL, ["__NONE__", "INTERPOLATE_WO", "INTERPOLATE_WO", "sgof", 2]),
        (EXAMPLE_SCAL, ["__NONE__", "INTERPOLATE_WO", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__BASE__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__LOW__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__PESS__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__PESSIMISTIC__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__OPT__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__OPTIMISTIC__", "__NONE__", "sgof", 1]),
        (EXAMPLE_SCAL, ["__NONE__", "__BASE__", "__OPTIMISTIC__", "sgof", 1]),
    ],
)
def test_fm_pyscal(dframe, runargs, tmpdir):
    """Parametrized test function for fm_pyscal"""
    tmpdir.chdir()
    dframe.to_csv("relperm-input.csv", index=False)

    # Insert a genkw-prefix in some parameters.txt files:
    if random.randint(0, 1):
        genkw_prefix = "FOO:"
    else:
        genkw_prefix = ""

    if "CASE" in dframe:
        with open("parameters.txt", "w") as p_file:
            p_file.write("INTERPOLATE_WO 0.1\n" + genkw_prefix + "INTERPOLATE_GO 0.5")

    run(*(["relperm-input.csv", "relperm.inc"] + runargs))
    assert os.path.exists("relperm.inc")
    assert len(open("relperm.inc").readlines()) > 20


def test_fm_pysal_static_xlsx(tmpdir):
    """Test fm_pyscal on a static xlsx file"""
    tmpdir.chdir()
    EXAMPLE_STATIC_DFRAME.to_excel("relperm-input.xlsx")
    run("relperm-input.xlsx", "relperm.inc", "", "__NONE__", "__NONE__", "slgof", 1)

    # pylint: disable=abstract-class-instantiated
    with pd.ExcelWriter("relperm-sheets.xlsx") as writer:
        EXAMPLE_STATIC_DFRAME.to_excel(writer, sheet_name="static")
        EXAMPLE_WATEROIL.to_excel(writer, sheet_name="wateroil")
    run(
        "relperm-sheets.xlsx", "static.inc", "static", "__NONE__", "__NONE__", "sgof", 1
    )
    assert os.path.exists("static.inc")
    assert "SGOF" in "".join(open("static.inc").readlines())
    run(
        "relperm-sheets.xlsx",
        "wateroil.inc",
        "wateroil",
        "__NONE__",
        "__NONE__",
        "sgof",
        1,
    )
    assert os.path.exists("wateroil.inc")
    assert "SGOF" not in "".join(open("wateroil.inc").readlines())


def test_fm_pyscal_argparse(tmpdir):
    """Test the command line wrapper"""
    tmpdir.chdir()
    EXAMPLE_STATIC_DFRAME.to_excel("relperm-input.xlsx")
    sys.argv = [
        "fm_pyscal",
        "relperm-input.xlsx",
        "relperm.inc",
        "",
        "__NONE__",
        "__NONE__",
        "slgof",
        "1",
    ]
    main_entry_point()
    assert os.path.exists("relperm.inc")


@pytest.mark.parametrize(
    "raises, dframe, runargs",
    [
        (
            # Writing to a file we don't have permission to write to
            (OSError, IOError),
            EXAMPLE_STATIC_DFRAME,
            [
                "relperm-input.csv",
                "/relperm.inc",
                "",
                "__NONE__",
                "__NONE__",
                "sgof",
                1,
            ],
        ),
        (
            # Reading a file that does not exist
            (ValueError, SystemExit),
            None,
            ["not-existing.csv", "relperm.inc", "", "__NONE__", "__NONE__", "sgof", 1],
        ),
        (
            # Specifying a sheet that does not exist
            (Exception, SystemExit),
            # Pyscal 0.4.0 raises an UnboundLocalError in this case, but with correct
            # error messages upfront. The exact exception might improve, so just assert
            # anything is thrown:
            EXAMPLE_STATIC_DFRAME,
            ["file.xlsx", "rel.inc", "foo_sheet", "__NONE__", "__NONE__", "sgof", 1],
        ),
        (
            # GasOil-interpolation without WaterOil, this is not supported:
            (ValueError, SystemExit),
            EXAMPLE_SCAL,
            [
                "scal-input.csv",
                "relperm.inc",
                None,
                "__NONE__",
                "INTERPOLATE_GO",
                "sgof",
                1,
            ],
        ),
        (
            # Non-existing parameter name
            (ValueError, SystemExit),
            EXAMPLE_SCAL,
            [
                "scal-input.csv",
                "relperm.inc",
                None,
                "FOOOO",
                "INTERPOLATE_GO",
                "sgof",
                1,
            ],
        ),
        (
            # Non-existing parameter name
            (ValueError, SystemExit),
            EXAMPLE_SCAL,
            [
                "scal-input.csv",
                "relperm.inc",
                None,
                "INTERPOLATE_WO",
                "INTERPOLATE_GOOOOOO",
                "sgof",
                1,
            ],
        ),
        (
            # Wrong magic name for interpolation parameter
            (ValueError, SystemExit),
            EXAMPLE_SCAL,
            [
                "scal-input.csv",
                "relperm.inc",
                None,
                "__OPTIM__",
                "__NONE__",
                "sgof",
                1,
            ],
        ),
        (
            # Wrong sgof type
            (ValueError, SystemExit),
            EXAMPLE_SCAL,
            [
                "scal-input.csv",
                "relperm.inc",
                None,
                "INTERPOLATE_WO",
                "INTERPOLATE_GO",
                "sgofffff",
                1,
            ],
        ),
        (
            # Wrong family type
            (ValueError, SystemExit),
            EXAMPLE_SCAL,
            [
                "scal-input.csv",
                "relperm.inc",
                None,
                "INTERPOLATE_WO",
                "INTERPOLATE_GO",
                "sgof",
                3,
            ],
        ),
    ],
)
def test_fm_pyscal_errors(raises, dframe, runargs, tmpdir):
    """Test that fm_pyscal or pyscal gives correct errors to
    common misusages"""
    tmpdir.chdir()
    if dframe is not None:
        if runargs[0].endswith("csv"):
            dframe.to_csv(runargs[0])
        else:
            dframe.to_excel(runargs[0])
    if dframe is not None and "CASE" in dframe:
        with open("parameters.txt", "w") as p_file:
            p_file.write("INTERPOLATE_WO 0.1\nINTERPOLATE_GO 0.5")

    with pytest.raises(raises):
        run(*runargs)



import numpy as np
import pandas as pd
import pytest

from semeio.workflows.correlated_observations_scaling.scaled_matrix import DataMatrix


def test_get_scaling_factor():
    np.random.seed(123)
    input_matrix = np.random.rand(10, 10)

    matrix = DataMatrix(pd.DataFrame(data=input_matrix))
    assert matrix.get_scaling_factor(6) == np.sqrt(10 / 6.0)


@pytest.mark.parametrize(
    "threshold,expected_result", [(0.0, 1), (0.83, 4), (0.90, 5), (0.95, 6), (0.99, 7)]
)
def test_get_nr_primary_components(threshold, expected_result):
    np.random.seed(123)
    input_matrix = np.random.rand(10, 10)
    matrix = DataMatrix(pd.DataFrame(data=input_matrix))
    components, _ = matrix.get_nr_primary_components(threshold)
    assert components == expected_result


def test_std_normalization():
    input_matrix = pd.DataFrame(np.ones((3, 3)))
    input_matrix.loc["OBS"] = np.ones(3)
    input_matrix.loc["STD"] = np.ones(3) * 0.1
    expected_matrix = [[10.0, 10.0, 10.0], [10.0, 10.0, 10.0], [10.0, 10.0, 10.0]]
    matrix = DataMatrix(pd.concat({"A_KEY": input_matrix}, axis=1))
    result = matrix.get_normalized_by_std()
    assert (result.loc[[0, 1, 2]].values == expected_matrix).all()



#!/usr/bin/env python

import argparse
import logging
import sys
import os

from semeio.jobs.rft.utility import (
    load_and_parse_well_time_file,
    valid_eclbase,
    existing_directory,
)
from semeio.jobs.rft import gendata_rft
from semeio.jobs.rft.trajectory import Trajectory
from semeio.jobs.rft.zonemap import ZoneMap

logger = logging.getLogger(__name__)

description = """
The gendata_rft application is used to retrieve rft data from an eclipse output
(or equivalent data format), given a file containing relevant well names and
their corresponding dates. It is also possible to specify a zonemap that validates
each zone specified in the trajectory files. See details of how each file should
be formatted for the individual arguments.

The user will be prompted with a warning and the applications stops if there is
no RFT available for the well at the date.

The pressure is set to -1 and the inactive value to 0 if any of the following applies:

* There is no RFT cell related to the trajectory point
* The trajectory point can not be found in the grid
* The zone mapping is invalid.

"""

category = "utility.transformation"


def _build_parser():
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "-e",
        "--eclbase",
        type=valid_eclbase,
        required=True,
        help="""
Filepath to an eclipse base (or with equivalent data format) that
contains both RFT and EGRID files. Note that the proper extensions are added,
so it is only necessary to specify the basename
""",
    )
    parser.add_argument(
        "-w",
        "--well_and_time_file",
        type=load_and_parse_well_time_file,
        required=True,
        help="""
Filepath to a file containing the well and time that will be loaded from the RFT
file. The accepted file format is: <well_name> <DD> <MM> <YY> <report>.
For each well name given in the file there must also exist a <well_name.txt> file
that contains a list of trajectory points. Each line must be specified according
to: <utmx> <utmy> <TVD> <MD>
""",
    )
    parser.add_argument(
        "-t",
        "--trajectory_path",
        required=True,
        type=existing_directory,
        help="""
Path to the directory that contains the trajectory files referenced from the well
and time file. Each well must have a trajectory file that contains at least one
trajectory point.
""",
    )
    parser.add_argument(
        "-z",
        "--zonemap",
        required=False,
        type=ZoneMap.load_and_parse_zonemap_file,
        help="""
Path to a file with zonemap information. This file must include every k-value that
any RFT value is located within, irregardless if the trajectory point contains a
zone definition. If a k value does not exist the program will stop. The zone
mapping will be validated and if it is invalid, the pressure is set to -1 and the
inactive value to 0
""",
    )
    parser.add_argument(
        "-c",
        "--csvfile",
        required=False,
        type=str,
        default="gendata_rft.csv",
        help="""
Path to a file to which a CSV output is dumped. The CSV will contain all the
information of the wellspaths, times, pressures and zone information, and whether
a specific pressure point is valid
""",
    )
    parser.add_argument(
        "-o",
        "--outputdirectory",
        required=False,
        type=existing_directory,
        default=".",
        help=(
            "Output directory used for outputting resulting files "
            "(except for the csvfile output). "
            "The directory must exist. "
            "Defaults to RUNPATH."
        ),
    )
    parser.add_argument(
        "--log-level",
        "-l",
        required=False,
        default="WARNING",
        type=logging.getLevelName,
        help="Sets the log level",
    )

    return parser


def main_entry_point(args=None):
    arg_parser = _build_parser()
    options = arg_parser.parse_args(args)
    logger.setLevel(options.log_level)

    well_names = [w_info[0] for w_info in options.well_and_time_file]

    trajectories = {
        wname: Trajectory.load_from_file(
            filepath=os.path.join(options.trajectory_path, wname + ".txt")
        )
        for wname in well_names
    }

    logger.info("All files loaded\nRetrieving RFT data...")

    try:
        gendata_rft.run(
            well_times=options.well_and_time_file,
            trajectories=trajectories,
            ecl_grid=options.eclbase[0],
            ecl_rft=options.eclbase[1],
            zonemap=options.zonemap,
            csvfile=options.csvfile,
            outputdirectory=options.outputdirectory,
        )
        with open("GENDATA_RFT.OK", "w") as fh:
            fh.write("GENDATA RFT completed OK")
        logger.info("Completed!")
    except ValueError as exception:
        logger.error("Failed with error message: {}".format(exception))


if __name__ == "__main__":
    main_entry_point(sys.argv[1:])



import collections

import yaml
import configsuite

from ert_data.measured import MeasuredData
from ert_shared.libres_facade import LibresFacade
from ert_shared.plugins.plugin_manager import hook_implementation
from semeio.communication import SemeioScript
from semeio.workflows.correlated_observations_scaling import job_config
from semeio.workflows.correlated_observations_scaling import ScalingJob
from semeio.workflows.correlated_observations_scaling.obs_utils import keys_with_data


_DESCRIPTION = """
    Correlated_observation_scaling is an ERT workflow job that does Principal
    Component Analysis (PCA) scaling of observations in ERT. The job accepts a
    configuration as the only argument.

    A valid job configuration is a YAML file in one of three formats:
    a CALCULATE_KEYS and UPDATE_KEYS pair; a list of such pairs; or a singular
    CALCULATE_KEYS directive. Each directive accepts a list of keys that
    each are a key and index hash. Keys can contain wildcards, and indices
    can represent ranges like "1-10,11,12". No index value implies all
    values. CALCULATE_KEYS can be configured using threshold which sets the
    threshold for PCA scaling [default: 0.95]; std_cutoff, the cutoff for
    standard deviation filtering [defaults to ``ert`` values]; and alpha for filtering
    between ensemble mean and observations [defaults to ``ert`` values].

Correlated Observations Scaling Configuration
=============================================
{}
""".format(
    configsuite.docs.generate(job_config.build_schema())
)


def _get_example(config_example):
    example_string = yaml.dump(config_example, default_flow_style=False)
    return "\n      ".join(example_string.split("\n"))


_calc_keys_example = {"CALCULATE_KEYS": {"keys": [{"key": "FOPR"}]}}
_calc_keys_w_index = {
    "CALCULATE_KEYS": {"keys": [{"key": "FOPR", "index": "1-10,50-100"}]}
}
_all_keys_w_index = {
    "CALCULATE_KEYS": {"keys": [{"key": "FOPR", "index": "1-10,50-100"}]},
    "UPDATE_KEYS": {"keys": [{"key": "FOPR", "index": "50-100"}]},
}
_wildcard_example = {"CALCULATE_KEYS": {"keys": [{"key": "WOPR_OP*"}]}}

_groups_example = [
    {"CALCULATE_KEYS": {"keys": [{"key": "FOPR"}]}},
    {"CALCULATE_KEYS": {"keys": [{"key": "WOPR_OP*"}]}},
]

_EXAMPLES = """

Configuration
-------------

Given a configuration in the format:

.. code-block:: yaml

    {calc_keys}

This configuration will let COS calculate a scaling factor from all data
points in ``FOPR`` and update all data points in ``FOPR``.
You can also specify which indices will be updated on the ``FOPR`` key:


Key indices
^^^^^^^^^^^

.. code-block:: yaml

    {calc_w_index}

This will calculate the scaling factor from indices 1-10 and 50-100, as
well as update these indices.

If not provided ``UPDATE_KEYS`` will use the same keys configuration as
``CALCULATE_KEYS``. Provided, it allows to specify which keys are to be
scaled:

.. code-block:: yaml

    {all_keys}

This configuration will calculate a scaling factor from indices 1-10,50-100
on ``FOPR``, but only update the scaling on indices ``50-100``.


Wildcards
^^^^^^^^^

Keys can be given as wildcards:

.. code-block:: yaml

    {wildcard}
This will calculate a scaling factor for all keys matching ``WOPR_OP*`` where
the asteriks will match everything.


Configuration for clusters
^^^^^^^^^^^^^^^^^^^^^^^^^^

For clusters (groups) of keys, a list of ``CALCULATE_KEYS`` and
``UPDATE_KEYS`` can be provided (the latter omitted for brevity):

.. code-block:: yaml

    {group}

This will calculate the scaling factor and do the scaling twice, instead
of passing two different configs.

""".format(
    calc_keys=_get_example(_calc_keys_example),
    calc_w_index=_get_example(_calc_keys_w_index),
    all_keys=_get_example(_all_keys_w_index),
    wildcard=_get_example(_wildcard_example),
    group=_get_example(_groups_example),
)


class CorrelatedObservationsScalingJob(
    SemeioScript
):  # pylint: disable=too-few-public-methods
    def run(self, job_config):
        facade = LibresFacade(self.ert())
        user_config = load_yaml(job_config)
        user_config = _insert_default_group(user_config)

        obs = facade.get_observations()
        obs_keys = [facade.get_observation_key(nr) for nr, _ in enumerate(obs)]
        obs_with_data = keys_with_data(
            obs,
            obs_keys,
            facade.get_ensemble_size(),
            facade.get_current_fs(),
        )
        default_values = _get_default_values(
            facade.get_alpha(), facade.get_std_cutoff()
        )
        for config in user_config:
            job = ScalingJob(
                obs_keys, obs, obs_with_data, config, self.reporter, default_values
            )
            measured_data = MeasuredData(
                facade, job.get_calc_keys(), job.get_index_lists()
            )
            job.scale(measured_data)


def load_yaml(job_config):
    # Allow job_config to be both list and dict.
    if isinstance(job_config, dict) or isinstance(job_config, list):
        return job_config

    with open(job_config, "r") as fin:
        return yaml.safe_load(fin)


def _insert_default_group(value):
    if isinstance(value, collections.Mapping):
        return [value]
    return value


def _get_default_values(alpha, std_cutoff):
    return {
        "CALCULATE_KEYS": {"std_cutoff": std_cutoff, "alpha": alpha},
        "UPDATE_KEYS": {},
    }


@hook_implementation
def legacy_ertscript_workflow(config):
    workflow = config.add_workflow(
        CorrelatedObservationsScalingJob, "CORRELATED_OBSERVATIONS_SCALING"
    )
    workflow.description = _DESCRIPTION
    workflow.examples = _EXAMPLES



from unittest.mock import patch
import pytest
from semeio.jobs.scripts.fm_stea import main_entry_point
import shutil
import os
from stea import SteaResult, SteaKeys


TEST_STEA_PATH, _ = os.path.split(os.path.abspath(__file__))


@pytest.fixture()
def setup_stea(tmpdir):
    cwd = os.getcwd()
    tmpdir.chdir()
    shutil.copytree(TEST_STEA_PATH, "stea")
    os.chdir(os.path.join("stea"))
    yield
    os.chdir(cwd)


def calculate_patch(stea_input):
    return SteaResult(
        {
            SteaKeys.KEY_VALUES: [
                {SteaKeys.TAX_MODE: SteaKeys.CORPORATE, SteaKeys.VALUES: {"NPV": 30}}
            ]
        },
        stea_input,
    )


@patch("stea.calculate")
@pytest.mark.usefixtures("setup_stea")
def test_stea(mock_stea):
    mock_stea.side_effect = calculate_patch
    main_entry_point(["-c", "stea_input.yml"])
    mock_stea.assert_called_once()
    files = os.listdir(os.getcwd())
    # the resulting file i.e. key is defined in the input config file: stea_input.yml
    assert "NPV_0" in files



from semeio.communication.reporter import FileReporter
from semeio.communication.semeio_script import SemeioScript, SEMEIOSCRIPT_LOG_FILE

__all__ = ("FileReporter", "SemeioScript", "SEMEIOSCRIPT_LOG_FILE")



# -*- coding: utf-8 -*-
import logging
import configsuite
from semeio.workflows.correlated_observations_scaling import job_config
from semeio.workflows.correlated_observations_scaling.exceptions import ValidationError
from semeio.workflows.correlated_observations_scaling.obs_utils import (
    create_active_lists,
    find_and_expand_wildcards,
)
from semeio.workflows.correlated_observations_scaling.scaled_matrix import DataMatrix
from semeio.workflows.correlated_observations_scaling.validator import (
    has_keys,
    is_subset,
)


class ScalingJob(object):
    def __init__(
        self,
        obs_keys,
        obs,
        obs_with_data,
        user_config_dict,
        reporter,
        default_values=None,
    ):
        """Creates a ScalingJob instance with the given obs_keys, obs,
        obs_with_data and user_config_dict."""
        if default_values is None:
            default_values = {}
        self._obs = obs
        self._obs_with_data = obs_with_data
        self._obs_keys = obs_keys
        self._config = self._setup_configuration(user_config_dict, default_values)
        self._validate()
        self._reporter = reporter

    def scale(self, measured_data):
        """
        Collects data, performs scaling and applies scaling, assumes validated input.
        """
        config = self._config.snapshot

        measured_data.remove_failed_realizations()
        measured_data.remove_inactive_observations()
        measured_data.filter_ensemble_mean_obs(config.CALCULATE_KEYS.alpha)
        measured_data.filter_ensemble_std(config.CALCULATE_KEYS.std_cutoff)

        matrix = DataMatrix(measured_data.data)
        matrix.normalize_by_std()

        events = config.CALCULATE_KEYS
        nr_components, singular_values = matrix.get_nr_primary_components(
            threshold=events.threshold
        )
        self._reporter.publish("svd", list(singular_values))

        logging.info("Scaling factor calculated from {}".format(events.keys))
        scale_factor = matrix.get_scaling_factor(nr_components)

        self._reporter.publish("scale_factor", scale_factor)

        update_data = create_active_lists(self._obs, config.UPDATE_KEYS.keys)
        self._update_scaling(self._obs, scale_factor, update_data)

    def _validate(self):
        """
        Validates the job. If invalid, raises an ValidationError.
        """
        errors = [] if self._config.valid else list(self._config.errors)
        calc_keys = self.get_calc_keys()
        application_keys = [
            entry.key for entry in self._config.snapshot.UPDATE_KEYS.keys
        ]
        errors.extend(is_subset(calc_keys, application_keys))
        errors.extend(
            has_keys(self._obs_keys, calc_keys, "Key: {} has no observations")
        )
        errors.extend(has_keys(self._obs_with_data, calc_keys, "Key: {} has no data"))
        if len(errors) > 0:
            raise ValidationError("Invalid job", errors)

    def get_calc_keys(self):
        return [event.key for event in self._config.snapshot.CALCULATE_KEYS.keys]

    def get_index_lists(self):
        return [
            event.index if len(event.index) > 0 else None
            for event in self._config.snapshot.CALCULATE_KEYS.keys
        ]

    def _setup_configuration(self, config_data, default_values):
        """
        Creates a ConfigSuite instance and inserts default values
        """
        schema = job_config.build_schema()
        config_dict = find_and_expand_wildcards(self._obs_keys, config_data)
        config = configsuite.ConfigSuite(
            config_dict,
            schema,
            deduce_required=True,
            layers=(default_values,),
        )
        return config

    @staticmethod
    def _update_scaling(obs, scale_factor, events):
        """
        Applies the scaling factor to the user specified index, SUMMARY_OBS needs to be
        treated differently as it only has one data point per node, compared with other
        observation types which have multiple data points per node.
        """
        for event in events:
            obs_vector = obs[event.key]
            for index, obs_node in enumerate(obs_vector):
                if obs_vector.getImplementationType().name == "SUMMARY_OBS":
                    index_list = (
                        event.index
                        if event.index is not None
                        else range(len(obs_vector))
                    )
                    if index in index_list:
                        obs_node.set_std_scaling(scale_factor)
                elif obs_vector.getImplementationType().name != "SUMMARY_OBS":
                    obs_node.updateStdScaling(scale_factor, event.active_list)
        logging.info(
            "Keys: {} scaled with scaling factor: {}".format(
                [event.key for event in events], scale_factor
            )
        )



#!/usr/bin/env python

import stea
import argparse
from semeio import valid_file

description = (
    "STEA is a powerful economic analysis tool used for complex economic "
    "analysis and portfolio optimization. STEA helps you analyze single "
    "projects, large and small portfolios and complex decision trees. "
    "As output, for each of the entries in the result section of the "
    "yaml config file, STEA will create result files "
    "ex: Res1_0, Res2_0, .. Res#_0"
)


category = "modeling.financial"


def _get_args_parser():

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "-c",
        "--config",
        help="STEA config file, yaml format required",
        type=valid_file,
    )
    return parser


def main_entry_point(args=None):
    parser = _get_args_parser()
    options = parser.parse_args(args)
    stea_input = stea.SteaInput([options.config])
    res = stea.calculate(stea_input)
    for res, value in res.results(stea.SteaKeys.CORPORATE).items():
        with open("{}_0".format(res), "w") as ofh:
            ofh.write("{}\n".format(value))


if __name__ == "__main__":
    main_entry_point()



# -*- coding: utf-8 -*-
from copy import deepcopy

import six

import configsuite
from configsuite import MetaKeys as MK
from configsuite import types


@configsuite.validator_msg("Minimum length of index list must be > 1 for PCA")
def _min_length(value):
    return len(value) > 1


@configsuite.validator_msg("Minimum value of index must be >= 0")
def _min_value(value):
    return value >= 0


_num_convert_msg = "Will go through the input and try to convert to list of int"


@configsuite.transformation_msg(_num_convert_msg)
def _to_int_list(value):
    value = deepcopy(value)
    if isinstance(value, six.integer_types):
        return [value]
    elif isinstance(value, (list, tuple)):
        value = ",".join([str(x) for x in value])
    return _realize_list(value)


@configsuite.transformation_msg("Convert ranges and singeltons into list")
def _realize_list(input_string):
    """If input_string is not a string, input_string will be returned. If input_string
    is a string it is assumed to contain comma separated elements. Each element is
    assumed to be either a singelton or a range. A singleton is a single number,
    i.e.  7 or 14. A range is a lower and upper element of the range separated by a
    single '-'. When provided with a string we will either return a list containing the
    union of all the singeltons and the ranges, or raise a TypeError or ValueError if
    it fails in the process. _realize_list('1,2,4-7,14-15') ->
    [1, 2, 4, 5, 6, 7, 14, 15]
    """
    real_list = []
    for elem in input_string.split(","):
        if elem.startswith("-"):
            raise ValueError(
                "Elements can not be negative, neither singletons nor in range,"
                " got: {}".format(elem)
            )
        bounds = elem.split("-")
        if len(bounds) == 1:
            real_list.append(int(elem))
        elif len(bounds) == 2:
            if elem.count("-") != 1:
                raise ValueError("Did expect single '-' in range")
            lower_bound = int(bounds[0])
            upper_bound = int(bounds[1]) + 1

            if lower_bound > upper_bound:
                err_msg = "Lower bound of range expected to be smaller then upper bound"
                raise ValueError(err_msg)

            real_list += range(lower_bound, upper_bound)
        else:
            raise ValueError("Expected at most one '-' in an element")

    return real_list


_num_convert_msg = "Create UPDATE_KEYS from CALCULATE_KEYS as it was not specified"


@configsuite.transformation_msg(_num_convert_msg)
def _expand_input(input_value):
    expanded_values = deepcopy(input_value)
    if "CALCULATE_KEYS" in expanded_values and "UPDATE_KEYS" not in expanded_values:
        expanded_values.update(
            {"UPDATE_KEYS": {"keys": expanded_values["CALCULATE_KEYS"]["keys"]}}
        )
    return expanded_values


@configsuite.validator_msg("Threshold must be higher than 0 and lower than 1")
def _min_max_value(value):
    return 0.0 < value < 1.0


@configsuite.validator_msg("keys must be provided for CALCULATE_KEYS")
def _CALCULATE_KEYS_key_not_empty_list(content):
    return len(content) > 0


_CALCULATE_KEYS_DESCRIPTION = """
The keys that will be used for calculating a scaling factor and update all
data points within said keys. Include the indexes under "index" in order to
update selected data points. Use UPDATE_KEYS if you would like the scaling
to be applied to other keys than the ones listed under CALCULATE_KEYS.
The configuration accepts a list of CALCULATE_KEYS, facilitating update on
multiple groups for a single run, instead of having to run multiple times.

NB: Between runs on clusters, "threshold", "std_cutoff" and "std_cutoff" are
reset.

Example:

.. code-block:: yaml

    CALCULATE_KEYS:
        keys:
            -
                key: FOPR
             index: 1-10,50-100

This will calculate the scaling factor from indices 1-10 and 50-100, as
well as update these indices.
"""

_UPDATE_KEYS_DESCRIPTION = """
Unless provided, the keys to be updated are the same as the ones in
CALCULATE_KEYS. UPDATE_KEYS can be used to specify different indexes and or
keys to apply the scaling factor.

.. code-block:: yaml

    CALCULATE_KEYS:
        keys:
            -
                key: FOPR
                index: 1-10,50-100
    UPDATE_KEYS:
        keys:
            -
                key: FOPR
                index: 50-100

This configuration will calculate a scaling factor from indices 1-10,50-100
on "FOPR", but only update the scaling on indices "50-100".
"""

_KEYS_SCHEMA = {
    MK.ElementValidators: (_CALCULATE_KEYS_key_not_empty_list,),
    MK.Type: types.List,
    MK.Content: {
        MK.Item: {
            MK.Type: types.NamedDict,
            MK.Content: {
                "key": {
                    MK.Type: types.String,
                    MK.Description: "Name of the key. An asterisk is accepted "
                    "as a suffix to expand all matching keywords (e.g. "
                    "WOPR* will include WOPR:OP1)",
                },
                "index": {
                    MK.Type: types.List,
                    MK.LayerTransformation: _to_int_list,
                    MK.Description: "Indexes matching the data points relevant "
                    "for update. Accepts single integer and ranges e.g. "
                    "(1,2,4-6,14-15) ->[1, 2, 4, 5, 6, 14, 15]",
                    MK.Content: {
                        MK.Item: {
                            MK.Type: types.Integer,
                            MK.ElementValidators: (_min_value,),
                        }
                    },
                },
            },
        }
    },
}


def build_schema():
    return {
        MK.Type: types.NamedDict,
        MK.Description: "Keys and index lists from all scaled keys",
        MK.LayerTransformation: _expand_input,
        MK.Content: {
            "CALCULATE_KEYS": {
                MK.Type: types.NamedDict,
                MK.Description: _CALCULATE_KEYS_DESCRIPTION,
                MK.Content: {
                    "keys": _KEYS_SCHEMA,
                    "threshold": {
                        MK.Type: types.Number,
                        MK.ElementValidators: (_min_max_value,),
                        MK.Description: "Threshold used when computing primary"
                        "components of the clusters.",
                        MK.Default: 0.95,
                    },
                    "std_cutoff": {
                        MK.Type: types.Number,
                        MK.Description: "A lower bound on the ensemble standard"
                        " deviation. All data points with insufficient variation"
                        " will be dropped. The value is defaulted to the value used"
                        " in the ert run, but if a value is given in the config, "
                        " that value will be used",
                    },
                    "alpha": {
                        MK.Type: types.Number,
                        MK.Description: "Scalar controlling the allowed distance"
                        "between ensemble mean and observation. In particular, "
                        "if: `abs(observed_value - ensemble_mean) > "
                        "alpha * (ensenmble_std + observed_std)` the data point "
                        "will be dropped. The value is defaulted to the value used "
                        "in the ert run, but if a value is given in the config, "
                        "that value will be used",
                    },
                },
            },
            "UPDATE_KEYS": {
                MK.Type: types.NamedDict,
                MK.Description: _UPDATE_KEYS_DESCRIPTION,
                MK.Content: {"keys": _KEYS_SCHEMA},
            },
        },
    }



import filecmp
import logging
import os
from distutils.dir_util import copy_tree

import pandas as pd
import pytest

from semeio.jobs.design2params import design2params
from semeio.jobs.design_kw.design_kw import extract_key_value


@pytest.fixture
def input_data(tmpdir):
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
    copy_tree(data_dir, tmpdir.strpath)

    cwd = os.getcwd()
    tmpdir.chdir()

    yield

    os.chdir(cwd)


@pytest.mark.usefixtures("input_data")
@pytest.mark.parametrize(
    "test_file, expected_file",
    [
        (design2params._PARAMETERS_TXT, "refparameters.txt"),
        (design2params._DESIGN_MATRIX_TXT, "refdesignmatrix.txt"),
        (design2params._DESIGN_PARAMETERS_TXT, "refdesignparameters.txt"),
    ],
)
def test_run(test_file, expected_file):
    design2params.run(
        1,
        "design_matrix.xlsx",
        "DesignSheet01",
        "DefaultValues",
        design2params._PARAMETERS_TXT,
        log_level=logging.DEBUG,
    )

    assert filecmp.cmp(test_file, expected_file)

    with open(design2params._TARGET_FILE_TXT, "r") as status_file:
        status = status_file.read()
        assert status == "OK\n"


@pytest.mark.usefixtures("input_data")
@pytest.mark.parametrize(
    "test_file, expected_file",
    [
        ("new_parameters.txt", "refparameters_when_missing.txt"),
        (design2params._DESIGN_MATRIX_TXT, "refdesignmatrix.txt"),
        (design2params._DESIGN_PARAMETERS_TXT, "refparameters_when_missing.txt"),
    ],
)
def test_run_with_no_parameters_txt(test_file, expected_file):
    design2params.run(
        1,
        "design_matrix.xlsx",
        "DesignSheet01",
        "DefaultValues",
        "new_parameters.txt",
        log_level=logging.DEBUG,
    )

    assert filecmp.cmp(test_file, expected_file)

    with open(design2params._TARGET_FILE_TXT, "r") as status_file:
        status = status_file.read()
        assert status == "OK\n"


@pytest.mark.usefixtures("input_data")
@pytest.mark.parametrize(
    "test_file, expected_file",
    [
        (design2params._PARAMETERS_TXT, "refparameters_w_default.txt"),
        (design2params._DESIGN_MATRIX_TXT, "refdesignmatrix.txt"),
        (design2params._DESIGN_PARAMETERS_TXT, "refdesignparameters_w_default.txt"),
    ],
)
def test_run_with_default(test_file, expected_file):
    design2params.run(
        1,
        "design_matrix_tampered_defaults.xlsx",
        "DesignSheet01",
        "DefaultValues",
        design2params._PARAMETERS_TXT,
        log_level=logging.DEBUG,
    )

    assert filecmp.cmp(test_file, expected_file)

    with open(design2params._TARGET_FILE_TXT, "r") as status_file:
        status = status_file.read()
        assert status == "OK\n"


@pytest.mark.usefixtures("input_data")
@pytest.mark.parametrize(
    "test_file, expected_file",
    [
        (design2params._PARAMETERS_TXT, "refparameters_w_spaces.txt"),
        (design2params._DESIGN_MATRIX_TXT, "refdesignmatrix_w_spaces.txt"),
        (design2params._DESIGN_PARAMETERS_TXT, "refdesignparameters_w_spaces.txt"),
    ],
)
def test_run_with_spaces_in_cells(test_file, expected_file):
    design2params.run(
        1,
        "design_matrix_cells_with_spaces.xlsx",
        "DesignSheet01",
        "DefaultValues",
        design2params._PARAMETERS_TXT,
        log_level=logging.DEBUG,
    )

    assert filecmp.cmp(test_file, expected_file)

    with open(design2params._TARGET_FILE_TXT, "r") as status_file:
        status = status_file.read()
        assert status == "OK\n"


@pytest.mark.usefixtures("input_data")
@pytest.mark.parametrize("realization_id", [(0), (1), (15), (24)])
def test_runs_different_reals_all_ok(realization_id):
    design2params.run(
        realization_id,
        "design_matrix_tampered_defaults.xlsx",
        "DesignSheet01",
        "DefaultValues",
        design2params._PARAMETERS_TXT,
        log_level=logging.DEBUG,
    )

    with open(design2params._TARGET_FILE_TXT, "r") as status_file:
        status = status_file.read()
        assert status == "OK\n"


def test_empty_defaults(tmpdir):
    # pylint: disable=abstract-class-instantiated
    tmpdir.chdir()
    designsheet_df = pd.DataFrame()
    defaultssheet_df = pd.DataFrame()
    writer = pd.ExcelWriter("design_matrix.xlsx")
    designsheet_df.to_excel(writer, sheet_name="DesignSheet01", index=False)
    defaultssheet_df.to_excel(
        writer, sheet_name="DefaultValues", index=False, header=None
    )
    writer.save()

    parsed_defaults = design2params._read_defaultssheet(
        "design_matrix.xlsx", "DefaultValues"
    )
    assert parsed_defaults.empty
    assert (parsed_defaults.columns == ["keys", "defaults"]).all()


def test_one_column_defaults(tmpdir):
    # pylint: disable=abstract-class-instantiated
    tmpdir.chdir()
    designsheet_df = pd.DataFrame()
    defaultssheet_df = pd.DataFrame(data=[["foo"]])
    writer = pd.ExcelWriter("design_matrix.xlsx")
    designsheet_df.to_excel(
        writer,
        sheet_name="DesignSheet01",
        index=False,
    )
    defaultssheet_df.to_excel(
        writer, sheet_name="DefaultValues", index=False, header=None
    )
    writer.save()

    with pytest.raises(SystemExit):
        design2params._read_defaultssheet("design_matrix.xlsx", "DefaultValues")


def test_three_column_defaults(tmpdir):
    """
    A deprecation warning will be emitted in this test.

    Change this test later to assert the code will fail hard.
    """
    # pylint: disable=abstract-class-instantiated
    tmpdir.chdir()
    designsheet_df = pd.DataFrame()
    defaultssheet_df = pd.DataFrame(data=[["foo", "bar", "com"]])
    writer = pd.ExcelWriter("design_matrix.xlsx")
    designsheet_df.to_excel(writer, sheet_name="DesignSheet01", index=False)
    defaultssheet_df.to_excel(
        writer, sheet_name="DefaultValues", index=False, header=None
    )
    writer.save()

    parsed_defaults = design2params._read_defaultssheet(
        "design_matrix.xlsx", "DefaultValues"
    )
    assert len(parsed_defaults) == 1
    assert (parsed_defaults.columns == ["keys", "defaults"]).all()


def test_run_realization_not_exist(input_data):
    with pytest.raises(SystemExit):
        design2params.run(
            1000,
            "design_matrix.xlsx",
            "DesignSheet01",
            "DefaultValues",
            design2params._PARAMETERS_TXT,
            log_level=logging.DEBUG,
        )


def test_open_excel_file_not_existing(input_data):
    with pytest.raises(SystemExit):
        design2params._read_excel("not_existing.xlsx", "DesignSheet01")


def test_open_excel_file_not_xls(input_data):
    with pytest.raises(SystemExit):
        design2params._read_excel("parameters.txt", "DesignSheet01")


def test_open_excel_file_header_missing(input_data):
    with pytest.raises(SystemExit):
        design2params.run(
            1,
            "design_matrix_missing_header.xlsx",
            "DesignSheet01",
            "DefaultValues",
            design2params._PARAMETERS_TXT,
            log_level=logging.DEBUG,
        )


def test_open_excel_file_value_missing(input_data):
    with pytest.raises(SystemExit):
        design2params.run(
            1,
            "design_matrix_missing_value.xlsx",
            "DesignSheet01",
            "DefaultValues",
            design2params._PARAMETERS_TXT,
            log_level=logging.DEBUG,
        )


def test_open_excel_file_wrong_defaults(input_data):
    with pytest.raises(SystemExit):
        design2params.run(
            1,
            "design_matrix_missing_value.xlsx",
            "DesignSheet01",
            "DefaultValues",
            design2params._PARAMETERS_TXT,
            log_level=logging.DEBUG,
        )


@pytest.mark.parametrize(
    "cellvalue, expected_parameters_str",
    [
        ("TRUE", "TRUE"),
        ("FALSE", "FALSE"),
        ("True", "True"),
        (True, "True"),
        (False, "False"),
        (0, "0"),
        (0.000, "0"),
        (1, "1"),
        ("1", "1"),
        ("1.0", "1.0"),
        (1.0, "1"),  # This difference is related to to_excel()
        ("æøå", "æøå"),
        (1e-1, "0.1"),
        (1e-4, "0.0001"),
        (1e-5, "1e-05"),
        (1e-16, "1e-16"),
        (1e3, "1000"),
        (1e10, "10000000000"),
    ],
)
def test_single_cell_values(cellvalue, expected_parameters_str, tmpdir):
    """Test how certain single values go through the Excel input sheets
    and all the way to parameters.txt"""
    # pylint: disable=abstract-class-instantiated
    tmpdir.chdir()

    designsheet_df = pd.DataFrame(columns=["REAL", "SOMEKEY"], data=[[0, cellvalue]])

    defaultssheet_df = pd.DataFrame()
    writer = pd.ExcelWriter("design_matrix.xlsx")
    designsheet_df.to_excel(writer, sheet_name="DesignSheet01", index=False)
    defaultssheet_df.to_excel(
        writer, sheet_name="DefaultValues", index=False, header=None
    )
    writer.save()
    params_file = "parameters.txt"
    design2params.run(
        0,
        "design_matrix.xlsx",
        "DesignSheet01",
        "DefaultValues",
        params_file,
        log_level=logging.DEBUG,
    )
    with open(params_file) as p_file:
        params_lines = p_file.readlines()
    key_vals = extract_key_value(params_lines)
    assert key_vals["SOMEKEY"] == expected_parameters_str


@pytest.mark.parametrize(
    "cellvalues, expected_parameters_strs",
    [
        (["TRUE", "True"], ["TRUE", "True"]),
        (["TRUE", 1], ["TRUE", "1"]),
        (["True", 1.0], ["True", "1"]),
        (["True", "1.0"], ["True", "1.0"]),
        ([0, "1.0"], ["0", "1.0"]),
        ([1, 1.2], ["1", "1.2"]),
        (["æ", 1e10], ["æ", "10000000000"]),
    ],
)
def test_pair_cell_values(cellvalues, expected_parameters_strs, tmpdir):
    """Test how a pair of values, one for each realization, go through
    the Excel input sheets and all the way to parameters.txt.

    This is to ensure that differing datatypes in a single Excel columns does
    not affect individual values."""
    # pylint: disable=abstract-class-instantiated
    tmpdir.chdir()

    designsheet_df = pd.DataFrame(
        columns=["REAL", "SOMEKEY"], data=[[0, cellvalues[0]], [1, cellvalues[1]]]
    )

    defaultssheet_df = pd.DataFrame()
    writer = pd.ExcelWriter("design_matrix.xlsx")
    designsheet_df.to_excel(writer, sheet_name="DesignSheet01", index=False)
    defaultssheet_df.to_excel(
        writer, sheet_name="DefaultValues", index=False, header=None
    )
    writer.save()
    params_0 = "parameters0.txt"
    design2params.run(
        0,
        "design_matrix.xlsx",
        "DesignSheet01",
        "DefaultValues",
        params_0,
        log_level=logging.DEBUG,
    )
    params_1 = "parameters1.txt"
    design2params.run(
        1,
        "design_matrix.xlsx",
        "DesignSheet01",
        "DefaultValues",
        params_1,
        log_level=logging.DEBUG,
    )
    with open(params_0) as p_file:
        params_lines_0 = p_file.readlines()
    with open(params_1) as p_file:
        params_lines_1 = p_file.readlines()
    key_vals_real0 = extract_key_value(params_lines_0)
    key_vals_real1 = extract_key_value(params_lines_1)
    assert key_vals_real0["SOMEKEY"] == expected_parameters_strs[0]
    assert key_vals_real1["SOMEKEY"] == expected_parameters_strs[1]



import numpy as np
import pandas as pd
import pytest
import os
from ert_data import measured
from semeio.workflows.correlated_observations_scaling import ScalingJob
from semeio.workflows.correlated_observations_scaling.exceptions import ValidationError
from semeio.communication.reporter import FileReporter


def test_filter_on_column_index():
    matrix = np.random.rand(10, 10)

    index_lists = [[0, 1], [1, 2, 3], [1, 2, 3, 4, 5]]
    for index_list in index_lists:
        result = measured.MeasuredData._filter_on_column_index(
            pd.DataFrame(matrix), index_list
        )
        assert result.shape == (10, len(index_list))

    with pytest.raises(IndexError):
        measured.MeasuredData._filter_on_column_index(pd.DataFrame(matrix), [11])


@pytest.mark.parametrize(
    "calc_key,app_key,obs_keys,obs_with_data,scaling_job_content",
    [
        (
            "KEY_1",
            "KEY_1",
            ["KEY_1"],
            ["KEY_1"],
            {"get_index_lists": [None], "get_calc_keys": ["KEY_1"]},
        ),
        (
            "KEY_1",
            "KEY_1",
            ["KEY_1", "KEY_2"],
            ["KEY_1"],
            {"get_index_lists": [None], "get_calc_keys": ["KEY_1"]},
        ),
    ],
)
@pytest.mark.usefixtures("setup_tmpdir")
def test_valid_job(
    calc_key,
    app_key,
    obs_keys,
    obs_with_data,
    scaling_job_content,
):
    user_config_dict = {
        "CALCULATE_KEYS": {"keys": [{"key": calc_key}]},
        "UPDATE_KEYS": {"keys": [{"key": app_key}]},
    }
    default_vals = {
        "CALCULATE_KEYS": {"std_cutoff": 1.0e-6, "alpha": 3},
        "UPDATE_KEYS": {},
    }
    reporter = FileReporter(os.path.realpath(os.getcwd()))
    job = ScalingJob(
        obs_keys, [], obs_with_data, user_config_dict, reporter, default_vals
    )
    assert job.get_index_lists() == scaling_job_content["get_index_lists"]
    assert job.get_calc_keys() == scaling_job_content["get_calc_keys"]


@pytest.mark.parametrize("input_key", ["alpha", "std_cutoff"])
def test_config_value_overwrites_default(input_key):
    user_config_dict = {
        "CALCULATE_KEYS": {"keys": [{"key": "a_key"}], "alpha": 2, "std_cutoff": 0.01},
    }
    default_vals = {
        "CALCULATE_KEYS": {"std_cutoff": 1.0e-6, "alpha": 3},
        "UPDATE_KEYS": {},
    }
    reporter = FileReporter(os.path.realpath(os.getcwd()))
    job = ScalingJob(["a_key"], [], ["a_key"], user_config_dict, reporter, default_vals)

    # Check that the value equals the config value
    assert (
        getattr(job._config.snapshot.CALCULATE_KEYS, input_key)
        == user_config_dict["CALCULATE_KEYS"][input_key]
    )
    # Check that different from default value
    assert (
        getattr(job._config.snapshot.CALCULATE_KEYS, input_key)
        != default_vals["CALCULATE_KEYS"][input_key]
    )


@pytest.mark.parametrize("alpha", [True, False])
@pytest.mark.parametrize(
    "calc_key,app_key,obs_keys,obs_with_data,errors",
    [
        (
            "KEY_1",
            "KEY_1",
            ["KEY_1", "KEY_2"],
            ["KEY_2"],
            ["Key: KEY_1 has no data"],
        ),
        (
            "not_in_list",
            "KEY_1",
            ["KEY_1"],
            ["KEY_1"],
            [
                "Update key: KEY_1 missing from calculate keys: ['not_in_list']",
                "Key: not_in_list has no observations",
                "Key: not_in_list has no data",
            ],
        ),
        (
            "KEY_1",
            "not_in_list",
            ["KEY_1"],
            ["KEY_1"],
            ["Update key: not_in_list missing from calculate keys: ['KEY_1']"],
        ),
        (
            "KEY_1",
            "KEY_1",
            [],
            ["KEY_1"],
            ["Key: KEY_1 has no observations"],
        ),
        ("KEY_1", "KEY_1", ["KEY_1"], [], ["Key: KEY_1 has no data"]),
    ],
)
@pytest.mark.usefixtures("setup_tmpdir")
def test_invalid_job(
    calc_key,
    app_key,
    obs_keys,
    obs_with_data,
    errors,
    alpha,
):
    user_config_dict = {
        "CALCULATE_KEYS": {"keys": [{"key": calc_key}], "std_cutoff": 0.001},
        "UPDATE_KEYS": {"keys": [{"key": app_key}]},
    }

    if alpha:
        user_config_dict["CALCULATE_KEYS"]["alpha"] = 3
    else:
        errors.insert(
            0,
            "MissingKeyError(msg=Missing key: alpha, "
            "key_path=('CALCULATE_KEYS',), layer=None)",
        )

    reporter = FileReporter(os.path.realpath(os.getcwd()))

    with pytest.raises(ValidationError) as exc_info:
        ScalingJob(obs_keys, [], obs_with_data, user_config_dict, reporter)
    assert [str(elem) for elem in exc_info.value.errors] == errors



import datetime
import argparse
import pytest

from semeio.jobs.scripts.gendata_rft import load_and_parse_well_time_file


@pytest.fixture()
def initdir(tmpdir):
    tmpdir.chdir()
    valid_data = """
WELL-NAME1 1 1 2000 2
WELL-NAME2 2 3 2001 3
"""
    tmpdir.join("valid_well_and_time.txt").write(valid_data)
    comment = "-- this is a comment"
    valid_data = comment + "\n" + valid_data + comment
    tmpdir.join("valid_well_and_time_with_comments.txt").write(valid_data)

    incorrect_number = """
WELL-NAME1 2
"""
    tmpdir.join("incorrect_number.txt").write(incorrect_number)

    incorrect_report_format = """
WELL-NAME1 1 1 2000 incorrect_report_step
"""
    tmpdir.join("incorrect_report_format.txt").write(incorrect_report_format)

    incorrect_date_format = """
WELL-NAME1 day month year 1
"""
    tmpdir.join("incorrect_date_format.txt").write(incorrect_date_format)


def test_load(initdir):
    expected_results = [
        ("WELL-NAME1", datetime.date(2000, 1, 1), 2),
        ("WELL-NAME2", datetime.date(2001, 3, 2), 3),
    ]
    for fname in [
        "valid_well_and_time.txt",
        "valid_well_and_time_with_comments.txt",
    ]:
        well_times = load_and_parse_well_time_file(fname)

        for (exp_wname, exp_wtime, exp_report), (wname, wtime, report) in zip(
            expected_results, well_times
        ):
            assert wname == exp_wname
            assert wtime == exp_wtime
            assert report == exp_report


def test_invalid_load(initdir):
    errors = [
        "Unexpected number of tokens: expected 5 got 2",
        "Unable to convert incorrect_report_step to int",
        "Unable to parse date, expected day month year got: day month year",
        "The path non_existing does not exist",
    ]

    fnames = [
        "incorrect_number.txt",
        "incorrect_report_format.txt",
        "incorrect_date_format.txt",
        "non_existing",
    ]

    for fname, error in zip(fnames, errors):
        with pytest.raises(argparse.ArgumentTypeError) as msgcontext:
            load_and_parse_well_time_file(fname)
        assert error in msgcontext.value.args[0]



#!/usr/bin/env python
import os
from glob import glob
from setuptools import setup, find_packages


def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


job_files = package_files("semeio/jobs/config_jobs") + package_files(
    "semeio/jobs/config_workflow_jobs"
)
scripts = glob("semeio/jobs/scripts/[!__]*.py")

setup(
    name="semeio",
    use_scm_version={"write_to": "semeio/version.py"},
    author="Software Innovation Bergen, Equinor ASA",
    author_email="fg_gpl@statoil.com",
    url="https://github.com/equinor/semeio",
    description="Jobs and workflow jobs for Ert.",
    packages=find_packages(include=["semeio*"], exclude=["semeio.jobs.scripts"]),
    entry_points={
        "ert": [
            "semeio_jobs = semeio.hook_implementations.jobs",
            "SpearmanCorrelation = semeio.workflows.spearman_correlation_job.spearman_correlation",  # noqa
            "MisfitPreprocessor = semeio.workflows.misfit_preprocessor.misfit_preprocessor",  # noqa
            "CorrelatedObsScaling = semeio.workflows.correlated_observations_scaling.cos",  # noqa
            "CsvExport2Job = semeio.jobs.scripts.csv_export2",
        ]
    },
    license="GPL-3.0",
    platforms="any",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=[
        "configsuite>=0.6",
        "numpy",
        "pandas",
        "scipy",
        "xlrd",
        "stea",
        "pyscal>=0.4.0",
        "fmu-ensemble",
    ],
    setup_requires=["setuptools_scm"],
    package_data={"": job_files},
    scripts=scripts,
    include_package_data=True,
)



import sys
import os
import shutil
import subprocess
import pytest

from ert_shared.plugins.plugin_manager import ErtPluginContext
import semeio.hook_implementations.jobs
import ert_shared.hook_implementations


@pytest.mark.skipif(sys.platform == "darwin", reason="Skip test for Mac OS")
@pytest.mark.parametrize(
    "nosim_command,data_input,data_expected",
    [
        (
            "INSERT_NOSIM",
            "RUNSPEC\n",
            "RUNSPEC\nNOSIM\n",
        ),
        (
            "INSERT_NOSIM",
            "RUNSPEC -- some comment about RUNSPEC and whatnot\n-- and more RUNSPEC\n",
            "RUNSPEC\nNOSIM\n-- and more RUNSPEC\n",
        ),
        (
            "INSERT_NOSIM",
            "RUNSPEC some comment about RUNSPEC and whatnot\n-- and more RUNSPEC\n",
            "RUNSPEC\nNOSIM\n-- and more RUNSPEC\n",
        ),
        (
            "REMOVE_NOSIM",
            "RUNSPEC\nNOSIM\n",
            "RUNSPEC\n",
        ),
        (
            "REMOVE_NOSIM",
            "RUNSPEC\nNOSIM\n-- and more RUNSPEC\n",
            "RUNSPEC\n-- and more RUNSPEC\n",
        ),
        (
            "REMOVE_NOSIM",
            (
                "RUNSPEC for some reason valid -- some comment RUNSPEC\n"
                "-- more\n"
                "NOSIM   more valid comment because it is after 8 chars"
                " -- and now a valid comment\n"
                "-- and a comment about NOSIM as well, not to be touched\n"
            ),
            (
                "RUNSPEC for some reason valid -- some comment RUNSPEC\n"
                "-- more\n-- and a comment about NOSIM as well, not to be touched\n"
            ),
        ),
    ],
)
def test_nosim(setup_tmpdir, nosim_command, data_input, data_expected):
    shutil.copy(os.path.join(os.path.dirname(__file__), "data", "nosim.ert"), ".")

    with open("nosim.ert", "a") as fh:
        fh.writelines(["FORWARD_MODEL {}".format(nosim_command)])

    with open("TEST.DATA", "w") as fh:
        fh.write(data_input)

    with ErtPluginContext(
        plugins=[semeio.hook_implementations.jobs, ert_shared.hook_implementations]
    ):
        p = subprocess.Popen(
            ["ert", "test_run", "nosim.ert", "--verbose"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=os.environ.copy(),
        )
        p.wait()

    with open("nosim/real_0/iter_0/TEST.DATA") as fh:
        assert fh.read() == data_expected



import collections
import configsuite as cs
import copy
import fnmatch


_SCALING = "scaling"
_THRESHOLD = "threshold"
_STD_CUTOFF = "std_cutoff"
_ALPHA = "alpha"
_CLUSTERING = "clustering"
_METHOD = "method"
SPEARMAN_CORRELATION = "spearman_correlation"
_FCLUSTER = "fcluster"
_SPEARMAN_THRESHOLD = "t"
_CRITERION = "criterion"
_INCONSISTENT = "inconsistent"
_DISTANCE = "distance"
_MAXCLUST = "maxclust"
_MONOCRIT = "monocrit"
_MAXCLUST_MONOCRIT = "maxclust_monocrit"
_DEPTH = "depth"
_LINKAGE = "linkage"
_SINGLE = "single"
_METHODS = (
    _SINGLE,
    "complete",
    "average",
    "weighted",
    "centroid",
    "ward",
)
_METRIC = "metric"
_EUCLIDEAN = "euclidean"
_METRICS = (
    "braycurtis",
    "canberra",
    "chebyshev",
    "cityblock",
    "correlation",
    "cosine",
    "dice",
    _EUCLIDEAN,
    "hamming",
    "jaccard",
    "jensenshannon",
    "kulsinski",
    "mahalanobis",
    "matching",
    "minkowski",
    "rogerstanimoto",
    "russellrao",
    "seuclidean",
    "sokalmichener",
    "sokalsneath",
    "sqeuclidean",
    "yule",
)
_OBSERVATIONS = "observations"


_DEFAULT_VALUES = {
    _SCALING: {_THRESHOLD: 0.95},
    _OBSERVATIONS: (),
    _CLUSTERING: {
        _METHOD: SPEARMAN_CORRELATION,
        SPEARMAN_CORRELATION: {
            _FCLUSTER: {_CRITERION: _INCONSISTENT, _DEPTH: 2},
            _LINKAGE: {_METHOD: _SINGLE, _METRIC: _EUCLIDEAN},
        },
    },
}


def _bounds_validator(
    lower=None, lower_inclusive=True, upper=None, upper_inclusive=True
):
    msg = "Value must be within the interval {}, {}".format(
        "<"
        if lower is None
        else "{}{}".format(
            "[" if lower_inclusive else "(",
            lower,
        ),
        ">"
        if upper is None
        else "{}{}".format(
            upper,
            "]" if upper_inclusive else ")",
        ),
    )

    @cs.validator_msg(msg)
    def validator(value):
        lower_valid = lower is None or (
            lower <= value if lower_inclusive else lower < value
        )
        upper_valid = upper is None or (
            value <= upper if upper_inclusive else value < upper
        )
        return lower_valid and upper_valid

    return validator


def _one_of(*valid_values):
    msg = "Value must be one of: ({})".format(", ".join(map(str, valid_values)))

    @cs.validator_msg(msg)
    def validator(value):
        return value in valid_values

    return validator


@cs.validator_msg("t must be an integer if a max cluster criteria is choosen")
def _t_is_int_if_maxclust(spearman_config):
    if spearman_config[_CRITERION] in (_MAXCLUST, _MAXCLUST_MONOCRIT):
        return isinstance(spearman_config[_SPEARMAN_THRESHOLD], int)
    return True


@cs.transformation_msg(
    (
        "Will be defaulted to 5 for all max cluster criterias and "
        "1.15 for all other criterias"
    )
)
def _inject_default_t(config):
    config = copy.deepcopy(config)
    criterion = config.get(_CRITERION)
    if _SPEARMAN_THRESHOLD in config:
        pass
    elif criterion in (_MAXCLUST, _MAXCLUST_MONOCRIT):
        config[_SPEARMAN_THRESHOLD] = 5
    elif criterion in (_INCONSISTENT, _DISTANCE, _MONOCRIT):
        config[_SPEARMAN_THRESHOLD] = 1.15
    else:
        raise ValueError("Unknown criterion: {}".format(criterion))
    return config


SPEARMAN_CORRELATION_SCHEMA = {
    cs.MetaKeys.Type: cs.types.NamedDict,
    cs.MetaKeys.Description: (
        "The {sc} clustering is supporting multiple parameters."
    ).format(sc=SPEARMAN_CORRELATION),
    cs.MetaKeys.Content: {
        _FCLUSTER: {
            cs.MetaKeys.Type: cs.types.NamedDict,
            cs.MetaKeys.Description: (
                "The {fcluster} implementation is backed by scipy and for a "
                "more detailed description we refer the reader to the "
                "documentation of scipy.cluster.hierarchy.fcluster` "
                "(https://docs.scipy.org/doc/scipy/reference/generated"
                "/scipy.cluster.hierarchy.fcluster.html)."
            ).format(fcluster=_FCLUSTER),
            cs.MetaKeys.ElementValidators: (_t_is_int_if_maxclust,),
            cs.MetaKeys.Transformation: _inject_default_t,
            cs.MetaKeys.Content: {
                _SPEARMAN_THRESHOLD: {
                    cs.MetaKeys.Type: cs.types.Number,
                    cs.MetaKeys.Description: (
                        "Scalar threshold for the clustering. When a "
                        "'{maxclust}' {criterion} is used, the scalar gives "
                        "the maximum number of clusters to be formed."
                    ).format(maxclust=_MAXCLUST, criterion=_CRITERION),
                    cs.MetaKeys.ElementValidators: (
                        _bounds_validator(lower=0, lower_inclusive=False),
                    ),
                },
                _CRITERION: {
                    cs.MetaKeys.Type: cs.types.String,
                    cs.MetaKeys.Description: (
                        "The criterion to use in forming flat clusters. "
                        "Defaults to {default_criterion}."
                    ).format(
                        default_criterion=_DEFAULT_VALUES[_CLUSTERING][
                            SPEARMAN_CORRELATION
                        ][_FCLUSTER][_CRITERION]
                    ),
                    cs.MetaKeys.ElementValidators: (
                        _one_of(
                            _INCONSISTENT,
                            _DISTANCE,
                            _MAXCLUST,
                            _MONOCRIT,
                            _MAXCLUST_MONOCRIT,
                        ),
                    ),
                },
                _DEPTH: {
                    cs.MetaKeys.Type: cs.types.Integer,
                    cs.MetaKeys.Description: (
                        "The maximum depth to perform the {inconsistent} calculation. "
                        "Defaults to {default_depth}."
                    ).format(
                        inconsistent=_INCONSISTENT,
                        default_depth=_DEFAULT_VALUES[_CLUSTERING][
                            SPEARMAN_CORRELATION
                        ][_FCLUSTER][_DEPTH],
                    ),
                    cs.MetaKeys.ElementValidators: (_bounds_validator(lower=1),),
                },
            },
        },
        _LINKAGE: {
            cs.MetaKeys.Type: cs.types.NamedDict,
            cs.MetaKeys.Description: (
                "The {linkage} implementation is backed by scipy and for a "
                "more detailed description we refer the reader to the "
                "documentation of scipy.cluster.hierarchy.linkage` "
                "(https://docs.scipy.org/doc/scipy/reference/generated"
                "/scipy.cluster.hierarchy.linkage.html)."
            ).format(linkage=_LINKAGE),
            cs.MetaKeys.Content: {
                _METHOD: {
                    cs.MetaKeys.Type: cs.types.String,
                    cs.MetaKeys.Description: (
                        "Method used to calculate the distance between the clusters."
                    ),
                    cs.MetaKeys.ElementValidators: (_one_of(*_METHODS),),
                },
                _METRIC: {
                    cs.MetaKeys.Type: cs.types.String,
                    cs.MetaKeys.Description: (
                        "Distance metric used to calculate distances."
                    ),
                    cs.MetaKeys.ElementValidators: (_one_of(*_METRICS),),
                },
            },
        },
    },
}


_CLUSTERING_SCHEMA = {
    cs.MetaKeys.Type: cs.types.NamedDict,
    cs.MetaKeys.Description: (
        "The clustering supports tweaking multiple parameters if the initial "
        "results are not according to expectations."
    ),
    cs.MetaKeys.Content: {
        _METHOD: {
            cs.MetaKeys.Type: cs.types.String,
            cs.MetaKeys.Description: (
                "Currently the workflow only supports clustering by "
                "{spearman}.".format(spearman=SPEARMAN_CORRELATION)
            ),
            cs.MetaKeys.ElementValidators: (_one_of(SPEARMAN_CORRELATION),),
        },
        SPEARMAN_CORRELATION: SPEARMAN_CORRELATION_SCHEMA,
    },
}


_SCALING_SCHEMA = {
    cs.MetaKeys.Type: cs.types.NamedDict,
    cs.MetaKeys.Description: ("Computes primary components and scales them."),
    cs.MetaKeys.Content: {
        _THRESHOLD: {
            cs.MetaKeys.Type: cs.types.Number,
            cs.MetaKeys.Description: (
                "Threshold used when computing primary components of the "
                "clusters. Defaults to: {default}"
            ).format(default=_DEFAULT_VALUES[_SCALING][_THRESHOLD]),
            cs.MetaKeys.ElementValidators: (
                _bounds_validator(
                    lower=0,
                    lower_inclusive=False,
                    upper=1,
                    upper_inclusive=False,
                ),
            ),
        },
        _STD_CUTOFF: {
            cs.MetaKeys.Type: cs.types.Number,
            cs.MetaKeys.Description: (
                "A lower bound on the ensemble standard deviation. All data "
                "points with insufficient variation will be dropped."
            ),
            cs.MetaKeys.AllowNone: True,
            cs.MetaKeys.Required: False,
        },
        _ALPHA: {
            cs.MetaKeys.Type: cs.types.Number,
            cs.MetaKeys.Description: (
                "Scalar controlling the allowed distance between "
                "ensemble mean and observation. In particular, if: "
                "`abs(observed_value - ensemble_mean) "
                "> alpha * (ensenmble_std + observed_std)` "
                "the data point will be dropped."
            ),
            cs.MetaKeys.AllowNone: True,
            cs.MetaKeys.Required: False,
        },
    },
}


class _BooleanWithMessage:
    def __init__(self, value, msg):
        self._value = value
        self._msg = msg

    def __nonzero__(self):
        return self._value is True

    def __bool__(self):
        return self._value is True

    def __and__(self, other):
        return bool(self) and bool(other)

    @property
    def msg(self):
        return self._msg


# NB: This is a validator, not a transformation.
# This is a hack to avoid configsuite wrapping the return value
@cs.transformation_msg("Ensures that all requested observations are indeed present")
def _observations_present(observation_key, context):
    if observation_key in context.observation_keys:
        return _BooleanWithMessage(
            True,
            "Observation {} found".format(observation_key),
        )
    else:
        return _BooleanWithMessage(
            False,
            "Found no match for observation {}".format(observation_key),
        )


@cs.transformation_msg("Inject all as default and expand filters")
def _realize_filters(observation_keys, context):
    all_keys = tuple(context.observation_keys)
    if len(observation_keys) == 0:
        observation_keys = ("*",)

    matches = set()
    for obs_filter in observation_keys:
        new_matches = set(fnmatch.filter(all_keys, obs_filter))
        if len(new_matches) == 0:
            new_matches = set((obs_filter,))
        matches = matches.union(new_matches)

    return tuple(matches)


_OBSERVATION_SCHEMA = {
    cs.MetaKeys.Type: cs.types.List,
    cs.MetaKeys.Description: (
        "By default all observations are clustered. If this is not desired "
        "one can provide a list of observation names to cluster. Wildcards are "
        'supported. Example: ["OP_1_WWCT*", "SHUT_IN_OP1"]'
    ),
    cs.MetaKeys.ContextTransformation: _realize_filters,
    cs.MetaKeys.Content: {
        cs.MetaKeys.Item: {
            cs.MetaKeys.Type: cs.types.String,
            cs.MetaKeys.ContextValidators: (_observations_present,),
        }
    },
}


_SCHEMA = {
    cs.MetaKeys.Type: cs.types.NamedDict,
    cs.MetaKeys.Description: (
        "The Misfit Preprocessor workflow provides the users with "
        "tooling to cluster and scale observations to prevent overfitting to "
        "all or parts of the observations. Examples where overfitting is likely "
        "to happen are accumulative time series and when one have data sources "
        "with severely different sampling magnitudes. This workflow is intended "
        "to be an almost out-of-the-box solution, where the user can tweak some "
        "parameters to get reasonable clusterings. The rest will be configured "
        "according to whatever is at the time considered best practices."
    ),
    cs.MetaKeys.Content: {
        _OBSERVATIONS: _OBSERVATION_SCHEMA,
        _CLUSTERING: _CLUSTERING_SCHEMA,
        _SCALING: _SCALING_SCHEMA,
    },
}


class _ObservationContext:  # pylint: disable=too-few-public-methods
    def __init__(self, observation_keys):
        Context = collections.namedtuple("Context", ("observation_keys",))
        self._context = Context(set(observation_keys))

    def __call__(self, _):
        return self._context


class MisfitPreprocessorConfig(cs.ConfigSuite):
    def __init__(self, config_data, observation_keys):
        super().__init__(
            config_data,
            _SCHEMA,
            (_DEFAULT_VALUES,),
            extract_validation_context=_ObservationContext(observation_keys),
            extract_transformation_context=_ObservationContext(observation_keys),
        )


def assemble_config(misfit_preprocessor_config, measured_data):
    observation_names = set(col[0] for col in measured_data.data.columns)
    return MisfitPreprocessorConfig(misfit_preprocessor_config, observation_names)



import fnmatch
from collections import namedtuple
from copy import deepcopy

from ecl.util.util import BoolVector
from res.enkf import ActiveList, LocalObsdata, RealizationStateEnum
from semeio.workflows.correlated_observations_scaling.exceptions import ValidationError


def _wildcard_to_dict_list(matching_keys, entry):
    """
    One of either:
    entry = {"key": "WOPT*", "index": [1,2,3]}
    entry = {"key": "WOPT*"}
    """
    if "index" in entry:
        return [{"key": key, "index": entry["index"]} for key in matching_keys]
    else:
        return [{"key": key} for key in matching_keys]


def _expand_wildcard(obs_list, wildcard_key, entry):
    """
    Expands a wildcard
    """
    matching_keys = fnmatch.filter(obs_list, wildcard_key["key"])
    return _wildcard_to_dict_list(matching_keys, entry)


def find_and_expand_wildcards(obs_list, user_dict):
    """
    Loops through the user input and identifies wildcards in observation
    names and expands them.
    """
    new_dict = deepcopy(user_dict)
    for main_key, value in user_dict.items():
        new_entries = []
        non_matching_wildcards = []
        if main_key in ("UPDATE_KEYS", "CALCULATE_KEYS"):
            for val in value["keys"]:
                if "*" in val["key"]:
                    expanded = _expand_wildcard(obs_list, val, val["key"])
                    if len(expanded) == 0:
                        non_matching_wildcards.append(val["key"])
                    new_entries.extend(expanded)
                else:
                    new_entries.append(val)
            new_dict[main_key]["keys"] = new_entries

            if len(non_matching_wildcards) > 0:
                raise ValidationError(
                    "Invalid {}".format(main_key),
                    ["{} had no match".format(wc) for wc in non_matching_wildcards],
                )
    return new_dict


def create_active_lists(enkf_observations, events):
    """
    Will add observation vectors to observation data. Returns
    a list of tuples mirroring the user config but also containing
    the active list where the scaling factor will be applied.
    """
    new_events = []
    observation_data = LocalObsdata("some_name", enkf_observations)
    for event in events:
        observation_data.addObsVector(enkf_observations[event.key])

        obs_index = _data_index_to_obs_index(enkf_observations, event.key, event.index)
        new_active_list = _get_active_list(observation_data, event.key, obs_index)

        new_events.append(_make_tuple(event.key, event.index, new_active_list))

    return new_events


def _get_active_list(observation_data, key, index_list):
    """
    If the user doesn't supply an index list, the existing active
    list from the observation is used, otherwise an active list is
    created from the index list.
    """
    if index_list is not None:
        return _active_list_from_index_list(index_list)
    else:
        return observation_data.copy_active_list(key).setParent()


def _make_tuple(
    key,
    index,
    active_list,
    new_event=namedtuple("named_dict", ["key", "index", "active_list"]),
):
    return new_event(key, index, active_list)


def _active_list_from_index_list(index_list):
    """
    Creates an ActiveList from a list of indexes
    :param index_list: list of index
    :type index_list:  list
    :return: Active list, a c-object with mode (ALL-ACTIVE, PARTIALLY-ACTIVE, INACTIVE)
        and list of indices
    :rtype: active_list
    """
    active_list = ActiveList()
    for index in index_list:
        active_list.addActiveIndex(index)
    return active_list


def keys_with_data(observations, keys, ensamble_size, storage):
    """
    Checks that all keys have data and returns a list of error messages
    """
    active_realizations = storage.realizationList(RealizationStateEnum.STATE_HAS_DATA)

    if len(active_realizations) == 0:
        return []

    active_mask = BoolVector.createFromList(ensamble_size, active_realizations)
    return [key for key in keys if observations[key].hasData(active_mask, storage)]


def _data_index_to_obs_index(obs, obs_key, data_index_list):
    if obs[obs_key].getImplementationType().name != "GEN_OBS":
        return data_index_list
    elif data_index_list is None:
        return data_index_list

    for timestep in obs[obs_key].getStepList().asList():
        node = obs[obs_key].getNode(timestep)
        index_map = {node.getIndex(nr): nr for nr in range(len(node))}
    return [index_map[index] for index in data_index_list]



import logging
import shlex
import re


_STATUS_FILE_NAME = "DESIGN_KW.OK"

_logger = logging.getLogger(__name__)


def run(
    template_file_name,
    result_file_name,
    log_level,
    parameters_file_name="parameters.txt",
):
    # Get all key, value pairs
    # If FWL key is having multiple entries in the parameters file
    # KeyError is raised. This will be logged, and no OK
    # file is written

    _logger.setLevel(log_level)

    valid = True

    with open(parameters_file_name) as parameters_file:
        parameters = parameters_file.readlines()

    key_vals = extract_key_value(parameters)

    key_vals.update(rm_genkw_prefix(key_vals))

    with open(template_file_name, "r") as template_file:
        template = template_file.readlines()

    if valid:
        with open(result_file_name, "w") as result_file:
            for line in template:
                if not is_comment(line):
                    for key, value in key_vals.items():
                        line = line.replace("<{}>".format(key), str(value))

                    if not all_matched(line, template_file_name, template):
                        valid = False

                result_file.write(line)

    if valid:
        with open(_STATUS_FILE_NAME, "w") as status_file:
            status_file.write("DESIGN_KW OK\n")


def all_matched(line, template_file_name, template):
    valid = True
    for unmatched in unmatched_templates(line):
        if is_perl(template_file_name, template):
            _logger.warn(
                (
                    "{} not found in design matrix, but this is probably a Perl file"
                ).format(unmatched)
            )
        else:
            _logger.error("{} not found in design matrix".format(unmatched))
            valid = False
    return valid


def is_perl(file_name, template):
    return file_name.endswith(".pl") or template[0].find("perl") != -1


def unmatched_templates(line):
    bracketpattern = re.compile("<.+?>")
    if bracketpattern.search(line):
        return bracketpattern.findall(line)
    else:
        return []


def is_comment(line):
    ecl_comment_pattern = re.compile("^--")
    std_comment_pattern = re.compile("^#")
    return ecl_comment_pattern.search(line) or std_comment_pattern.search(line)


def extract_key_value(parameters):
    """Parses a list of strings, looking for key-value pairs pr. line
    separated by whitespace, into a dictionary.

    Spaces in keys and/or values are supported if quoted. Quotes
    in keys/values are not supported.

    Args:
        parameters (list of str)

    Returns:
        dict, with the keys and values parsed.

    Raises:
        ValueError, with error messages and all unparseable lines.
    """
    res = {}
    errors = []
    for line in parameters:
        line_parts = shlex.split(line)
        if not line_parts:
            continue
        if len(line_parts) == 1:
            errors += ["No value found in line {}".format(line)]
            continue
        if len(line_parts) > 2:
            errors += ["Too many values found in line {}".format(line)]
            continue
        key, value = line_parts
        if key in res:
            errors += ["{} is defined multiple times".format(key)]
            continue
        res[key] = value
    if errors:
        raise ValueError("\n".join(errors))
    return res


def rm_genkw_prefix(paramsdict, ignoreprefixes="LOG10_"):
    """Strip prefixes from keys in a dictionary.

    Prefix is any string before a colon. No colon means no prefix.

    Only keys unique after prefix-stripping
    are included. For intentional duplicates, as when ERT
    prepares LOG10_ values, these are ignored by default in this
    function.

    Args:
        paramsdict (dict): Dictionary with parameter names as keys.
        ignoreprefixes (str or list of str): If any of these strings
            are found at the start of the prefix, they are removed
            from the dictionary before uniqueness is determined.

    Returns:
        Subset of the incoming dictionary (ignored keys are dropped), and with
        stripped prefixes from keys.
    """
    if ignoreprefixes is None:
        ignoreprefixes = []
    if isinstance(ignoreprefixes, str):
        ignoreprefixes = [ignoreprefixes]
    ignoreprefixes = filter(None, ignoreprefixes)

    for ignore_str in ignoreprefixes:
        paramsdict = {
            key: paramsdict[key]
            for key in paramsdict
            if ":" not in key or not key.startswith(ignore_str)
        }

    keyvalues = [
        (key.split(":")[1], value) if ":" in key else (key, value)
        for key, value in paramsdict.items()
    ]

    keys = [keyval[0] for keyval in keyvalues]

    duplicates = {keyvalue[0] for keyvalue in keyvalues if keys.count(keyvalue[0]) > 1}
    if duplicates:
        _logger.warning(
            "Key(s) {} can only be used with prefix.".format(list(duplicates))
        )

    return {
        keyvalue[0]: keyvalue[1]
        for keyvalue in keyvalues
        if keys.count(keyvalue[0]) == 1
    }



import logging
from threading import Thread

from semeio.communication import SemeioScript


class TestWorkflowJob(SemeioScript):
    def run(self, *args):
        self.reporter.publish("test_data", list(range(10)))

        # The mission of this code is to simulate that something outside the
        # thread of the workflow (e.g. ERT itself), is logging. This should not
        # end up in the workflow log itself.
        thread = Thread(
            target=lambda: logging.error(
                "Log statement from outside the workflow thread."
            )
        )
        thread.start()
        thread.join()

        logging.error(
            "I finished without any problems - hence I'm not a failure after all!"
        )



import argparse
import os

import pandas as pd

from semeio.jobs.rft.utility import strip_comments


class TrajectoryPoint:
    """Represents a point along a wellpath. The point is
    determined by UTM x and y, measured depth along the wellpath, and
    true vertical depth.

    Points can be *active*, which means that (i,j,k) in a given grid
    is determined, there is simulated pressure available in that cell,
    and the cell in in the correct zone (mostly important for long
    horizontals).

    RKB for MD must match MD in Eclipse RFT data.
    """

    def __init__(self, utm_x, utm_y, measured_depth, true_vertical_depth, zone=None):
        self.utm_x = utm_x
        self.utm_y = utm_y
        self.measured_depth = measured_depth
        self.true_vertical_depth = true_vertical_depth
        self.zone = zone
        self.grid_ijk = None
        self.pressure = None
        self.valid_zone = False

    def set_ijk(self, point):
        self.grid_ijk = point

    def validate_zone(self, zonemap=None):
        if self.zone is None:
            self.valid_zone = True
        elif self.grid_ijk:
            self.valid_zone = zonemap is None or zonemap.has_relationship(
                self.zone, self.grid_ijk[2]
            )

    def is_active(self):
        return (
            self.grid_ijk is not None and self.pressure is not None and self.valid_zone
        )

    def inactive_info(self, zonemap=None):
        if self.grid_ijk is None:
            return "TRAJECTORY_POINT_NOT_IN_GRID {}".format(str(self))
        if self.pressure is None:
            return "TRAJECTORY_POINT_NOT_IN_RFT {}".format(str(self))
        if not self.valid_zone and zonemap is not None:
            if self.grid_ijk[2] not in zonemap:
                return "ZONEMAP_MISSING_VALUE {} {} {}".format(
                    str(self), self.grid_ijk[2], None
                )

            return "ZONE_MISMATCH {} {} {}".format(
                str(self), self.grid_ijk[2], zonemap[self.grid_ijk[2]]
            )

    def get_pressure(self):
        if self.is_active():
            return self.pressure
        return -1

    def update_pressure_from_rft(self, rft):
        if self.grid_ijk:
            rftcell = rft.ijkget(self.grid_ijk)
            if rftcell:
                self.pressure = rftcell.pressure

    def __str__(self):
        return "(utm_x={utm_x}, utm_y={utm_y}, measured_depth={measured_depth})".format(
            utm_x=self.utm_x, utm_y=self.utm_y, measured_depth=self.measured_depth
        )


class Trajectory:
    """Represents a well trajectory as a list of TrajectoryPoints.

    Args:
        points (list): List of lists with data for the trajectory.
    """

    def __init__(self, points):
        self.trajectory_points = [TrajectoryPoint(*point) for point in points]

    def __getitem__(self, i):
        return self.trajectory_points[i]

    def __len__(self):
        return len(self.trajectory_points)

    def to_dataframe(self, zonemap=None):
        """Expose the trajectory data as a Pandas DataFrame.

        The data grid_ijk in self is a tuple, this is split into
        three distinct columns, i, j and k.

        The dataframe is sorted first by measured_depth, if not available
        it is sorted by true_vertical_depth. The original order from the
        input files is conserved in the column "order".

        Returns:
            pd.DataFrame
        """
        dframe = pd.DataFrame(data=[vars(point) for point in self.trajectory_points])
        dframe["is_active"] = [point.is_active() for point in self.trajectory_points]
        dframe["inactive_info"] = [
            point.inactive_info(zonemap) for point in self.trajectory_points
        ]

        dframe = Trajectory.split_tuple_column(
            dframe, tuplecolumn="grid_ijk", components=["i", "j", "k"]
        )

        # Conserve original order before sorting - original order from
        # well input files is needed by some export routines.
        dframe.index.name = "order"
        dframe = dframe.reset_index()

        if "measured_depth" in dframe:
            dframe = dframe.sort_values("measured_depth")
        elif "true_vertical_depth" in dframe:
            dframe = dframe.sort_values("true_vertical_depth")

        return dframe.dropna(how="all", axis="columns")

    @staticmethod
    def split_tuple_column(dframe, tuplecolumn="grid_ijk", components=None):
        """For a dataframe with a column containing a tuple
        datatype, split the tuple into the components [I, J, K] as new columns.

        Args:
            dframe (pd.DataFrame): A dataframe with a tuple column
            tuplecolumn (str): Column name with tuples. Defaults "grid_ijk"
            components (list of str): Name of components in tuple, default
                ["i", "j", "k"]
        Returns:
            pd.DataFrame, one column removed, and 3 (if defaults) new columns added.
                Column order is undefined.
        """
        if components is None:
            components = ["i", "j", "k"]

        non_nulls = ~dframe[tuplecolumn].isnull()
        if not non_nulls.any():
            return dframe
        tuplelengths = {len(value) for value in dframe.loc[non_nulls, tuplecolumn]}
        if len(tuplelengths) != 1:
            raise ValueError(
                "Uneven tuple lengths {} in non-null dataframe data".format(
                    tuplelengths
                )
            )
        if list(tuplelengths)[0] != len(components):
            raise ValueError("Mismatch between tuple length and given column names")

        ijk_df = pd.DataFrame(
            data=dframe.loc[non_nulls, tuplecolumn].tolist(),
            index=dframe.index[non_nulls],
            columns=components,
        )
        return pd.concat(
            [dframe.drop(tuplecolumn, axis="columns"), ijk_df],
            axis="columns",
            ignore_index=False,
            sort=False,
        )

    @staticmethod
    def parse_trajectory_line(line):
        """Reads a text line with four floating point values (utm_x, utm_y, md, and tvd)
        and an optional string being the zone name.

        Returns:
            List of four floats or a list of four floats and a string.
        """
        point = line.split()
        if len(point) < 4 or len(point) > 5:
            raise argparse.ArgumentTypeError(
                (
                    "Trajectory data file not on correct format: "
                    "'utm_x utm_y md tvd <zone>' - zone is optional"
                )
            )

        try:
            floats = [float(v) for v in point[:4]]
        except ValueError:
            raise argparse.ArgumentTypeError(
                "Error: Failed to extract data from line {line}".format(line=line)
            )

        zone = point[4].strip() if len(point) == 5 else None
        return floats + [zone]

    @classmethod
    def load_from_file(cls, filepath):
        """Initialize a Trajectory from a text file describing a well path."""
        trajectory_points = []

        filename = os.path.join(filepath)
        if not os.path.isfile(filename):
            raise argparse.ArgumentTypeError(
                "Warning: Trajectory file {filename} not found!".format(
                    filename=filename
                )
            )

        with open(filename, "r") as f:
            trajectory_lines = f.readlines()

        trajectory_lines = [strip_comments(line) for line in trajectory_lines]

        for line in trajectory_lines:
            if not line:
                continue
            point = Trajectory.parse_trajectory_line(line)
            trajectory_points.append(point)

        return cls(trajectory_points)



import argparse
import sys

from res.enkf import ErtScript
from semeio.jobs.csv_export2 import csv_export2
from ert_shared.plugins.plugin_manager import hook_implementation


class CsvExport2Job(ErtScript):  # pylint: disable=too-few-public-methods
    def run(self, *args):
        main(args)


def main(args):
    parser = csv_export_parser()
    args = parser.parse_args(args)

    csv_export2.csv_exporter(
        runpathfile=args.runpathfile,
        time_index=args.time_index,
        outputfile=args.outputfile,
        column_keys=args.column_keys,
    )

    print("{} csv-export written to {}".format(args.time_index, args.outputfile))


def csv_export_parser():
    """Setup parser"""
    parser = argparse.ArgumentParser()
    parser.add_argument("runpathfile", type=str)
    parser.add_argument("outputfile", type=str)
    parser.add_argument("time_index", type=str, default="monthly")
    parser.add_argument("column_keys", nargs="+", default=None)
    return parser


if __name__ == "__main__":
    main(sys.argv[1:])


@hook_implementation
def legacy_ertscript_workflow(config):
    workflow = config.add_workflow(CsvExport2Job, "CSV_EXPORT2")
    workflow.parser = csv_export_parser



import logging
import os
from distutils.dir_util import copy_tree

import pytest

from semeio.jobs.scripts import design2params


@pytest.fixture
def input_data(tmpdir):
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
    copy_tree(data_dir, tmpdir.strpath)

    cwd = os.getcwd()
    tmpdir.chdir()

    yield

    os.chdir(cwd)


_realization = 1
_xls = "design_matrix.xlsx"
_sheet = "DesignSheet01"
_default_sheet = "DefaultSheet01"
_parametersfilename = "parameters.txt"
_log_level = "DEBUG"
_default_log_level = "WARNING"


def test_argparse(input_data):
    args = [str(_realization), _xls, _sheet, _default_sheet]
    parser = design2params.create_parser()
    res = parser.parse_args(args)
    assert res
    assert res.realization == _realization
    assert res.xlsfilename == _xls
    assert res.designsheetname == _sheet
    assert res.defaultssheetname == _default_sheet


def test_argparse_no_default(input_data):
    args = [str(_realization), _xls, _sheet]
    parser = design2params.create_parser()
    res = parser.parse_args(args)
    assert res
    assert res.realization == _realization
    assert res.xlsfilename == _xls
    assert res.designsheetname == _sheet
    assert res.defaultssheetname is None


def test_argparse_with_optionals(input_data):
    args = [
        str(_realization),
        _xls,
        _sheet,
        "-p",
        _parametersfilename,
        "-l",
        _log_level,
    ]
    parser = design2params.create_parser()
    res = parser.parse_args(args)
    assert res
    assert res.realization == _realization
    assert res.xlsfilename == _xls
    assert res.designsheetname == _sheet
    assert res.defaultssheetname is None
    assert res.parametersfilename == _parametersfilename
    assert res.log_level == logging.getLevelName(_log_level)


def test_argparse_xls_file_not_exists():
    with pytest.raises(SystemExit):
        design2params.main([str(_realization), "not_a_file", _sheet])


def test_argparse_parameters_file_not_exists(input_data):
    not_existing_parameters_txt = "not_a_file.txt"

    args = [
        str(_realization),
        _xls,
        _sheet,
        "-p",
        not_existing_parameters_txt,
    ]
    parser = design2params.create_parser()
    res = parser.parse_args(args)
    assert res
    assert res.realization == _realization
    assert res.xlsfilename == _xls
    assert res.designsheetname == _sheet
    assert res.defaultssheetname is None
    assert res.parametersfilename == not_existing_parameters_txt
    assert res.log_level == logging.getLevelName(_default_log_level)


def test_argparse_design_sheet_missing(input_data):
    with pytest.raises(SystemExit):
        design2params.main([str(_realization), _xls])



import importlib
import os
from pkg_resources import resource_filename

from ert_shared.plugins.plugin_manager import hook_implementation
from ert_shared.plugins.plugin_response import plugin_response


def _get_jobs_from_directory(directory):
    resource_directory = resource_filename("semeio", directory)

    all_files = [
        os.path.join(resource_directory, f)
        for f in os.listdir(resource_directory)
        if os.path.isfile(os.path.join(resource_directory, f))
    ]
    return {os.path.basename(path): path for path in all_files}


@hook_implementation
@plugin_response(plugin_name="semeio")  # pylint: disable=no-value-for-parameter
def installable_jobs():
    return _get_jobs_from_directory("jobs/config_jobs")


def _get_module_variable_if_exists(module_name, variable_name, default=""):
    try:
        script_module = importlib.import_module(module_name)
    except ImportError:
        return default

    return getattr(script_module, variable_name, default)


@hook_implementation
@plugin_response(plugin_name="semeio")  # pylint: disable=no-value-for-parameter
def job_documentation(job_name):
    semeio_jobs = set(installable_jobs().data.keys())
    if job_name not in semeio_jobs:
        return None

    if job_name.endswith("NOSIM"):
        verb = job_name.split("_")[0].lower()
        insert = verb == "insert"
        return {
            "description": (
                "{} NOSIM {} the ECLIPSE data file. "
                "This will {} simulation in ECLIPSE."
            ).format(
                verb.capitalize(),
                "into" if insert else "from",
                "disable" if insert else "enable",
            ),
            "examples": "",
            "category": "utility.eclipse",
        }

    if job_name == "STEA" or job_name == "PYSCAL":
        module_name = "semeio.jobs.scripts.fm_{}".format(job_name.lower())
    else:
        module_name = "semeio.jobs.scripts.{}".format(job_name.lower())

    description = _get_module_variable_if_exists(
        module_name=module_name, variable_name="description"
    )
    examples = _get_module_variable_if_exists(
        module_name=module_name, variable_name="examples"
    )
    category = _get_module_variable_if_exists(
        module_name=module_name, variable_name="category", default="other"
    )

    return {
        "description": description,
        "examples": examples,
        "category": category,
    }



#!/usr/bin/env python

import argparse
import logging
import sys

from semeio import valid_file
from semeio.jobs.design_kw import design_kw

description = """
Performs text -> value substitutions (similar to GEN_KW) in ``template_file``,
picking values from ``parameters.txt``.

This is designed for using in conjunction with a design matrix that have been
processed using ``DESIGN2PARAMS`` into ``parameters.txt``.

Fails hard if not templates can be replaced by a value.

The prefix (before the colon) in parameter names, as when written by ``GEN_KW``
and not ``DESIGN2PARAMS``, is optional in the template files.

Example: If ``parameters.txt`` has the line ``MULTFLT:FLT_A10_B18 0.001``, both
of the templates ``<MULTFLT:FLT_A10_B18>`` and ``<FLT_A10_B18>`` will expand to
``0.001``.
"""

category = "utility.templating"


def create_parser():
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "templatefile",
        type=valid_file,
        help="Path to template-file",
    )

    parser.add_argument(
        "resultfile",
        type=str,
        help="Path to result-file",
    )

    parser.add_argument(
        "--log-level",
        required=False,
        default="WARNING",
        type=logging.getLevelName,
    )

    return parser


def main(args):
    parser = create_parser()
    parsed_args = parser.parse_args(args)

    design_kw.run(
        template_file_name=parsed_args.templatefile,
        result_file_name=parsed_args.resultfile,
        log_level=parsed_args.log_level,
    )


if __name__ == "__main__":
    main(sys.argv[1:])



import json
import logging
import os
import pytest
from threading import Thread

from semeio.communication import SEMEIOSCRIPT_LOG_FILE, SemeioScript

from unittest.mock import Mock


def _ert_mock(config_path="config_path", user_config_file="config_file"):
    resconfig_mock = Mock()
    resconfig_mock.config_path = config_path
    resconfig_mock.user_config_file = user_config_file

    return Mock(resConfig=Mock(return_value=resconfig_mock))


def test_semeio_script_publish(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    namespace = "arbitrary_data"
    data = [1, 2, 3, 4, "This is a very important string"]

    class MySuperScript(SemeioScript):
        def run(self, *args):
            self.reporter.publish(namespace, data)

    my_super_script = MySuperScript(ert)
    my_super_script.run()

    expected_outputfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        namespace + ".json",
    )

    with open(expected_outputfile) as f:
        published_data = json.load(f)

    assert [data] == published_data


def test_semeio_script_logging(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    msg = "My logging msg"

    class MySuperScript(SemeioScript):
        def run(self, *args):
            logging.error(msg)

    my_super_script = MySuperScript(ert)
    my_super_script.run()

    expected_logfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    with open(expected_logfile) as f:
        loaded_log = f.readlines()

    assert len(loaded_log) == 1
    assert msg in loaded_log[0]


def assert_log(messages, log_file):
    with open(log_file) as f:
        log_data = f.readlines()

    assert len(messages) == len(log_data)
    for msg, log_entry in zip(messages, log_data):
        assert msg in log_entry


@pytest.mark.parametrize(
    "messages",
    [
        ("msg",),
        ("This is a message!", "And this is another one"),
        tuple("message_{}".format(idx) for idx in range(10)),
    ],
)
def test_semeio_script_multiple_logging(messages, tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    class MySuperScript(SemeioScript):
        def run(self, *args):
            posted_messages = []
            for msg in messages:
                logging.error(msg)
                posted_messages.append(msg)

                assert_log(posted_messages, expected_logfile)

    expected_logfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    my_super_script = MySuperScript(ert)
    my_super_script.run()

    assert_log(messages, expected_logfile)


def test_semeio_script_post_logging(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    class MySuperScript(SemeioScript):
        def run(self, *args):
            logging.error("A message from MySuperScript")

    my_super_script = MySuperScript(ert)
    my_super_script.run()

    logging.error("A second message - not from MySuperScript")

    expected_logfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    with open(expected_logfile) as f:
        loaded_log = f.readlines()

    assert len(loaded_log) == 1


def test_semeio_script_pre_logging(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    class MySuperScript(SemeioScript):
        def run(self, *args):
            logging.error("A message from MySuperScript")

    my_super_script = MySuperScript(ert)

    logging.error("A message - not from MySuperScript")

    my_super_script.run()

    expected_logfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    with open(expected_logfile) as f:
        loaded_log = f.readlines()

    assert len(loaded_log) == 1


def test_semeio_script_concurrent_logging(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    class MySuperScript(SemeioScript):
        def run(self, *args):
            logging.error("A message from MySuperScript")
            thread = Thread(target=lambda: logging.error("External event."))
            thread.start()
            thread.join()

    MySuperScript(ert).run()

    expected_logfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    with open(expected_logfile) as f:
        loaded_log = f.readlines()

    assert len(loaded_log) == 1


def test_semeio_script_post_logging_exception(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    class MySuperScript(SemeioScript):
        def run(self, *args):
            logging.error("A message from MySuperScript")
            raise AssertionError("Bazinga")

    my_super_script = MySuperScript(ert)
    try:
        my_super_script.run()
    except AssertionError:
        pass

    logging.error("A second message - not from MySuperScript")

    expected_logfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    with open(expected_logfile) as f:
        loaded_log = f.readlines()

    assert len(loaded_log) == 1


def test_semeio_script_keyword_args(tmpdir):
    tmpdir.chdir()
    config_path, user_config_file = "config_path", "config_file"
    ert = _ert_mock(config_path, user_config_file)

    class MySuperScript(SemeioScript):
        def run(self, param_A, param_B):
            self.reporter.publish_msg(SEMEIOSCRIPT_LOG_FILE, param_A)
            self.reporter.publish_msg(SEMEIOSCRIPT_LOG_FILE, param_B)

    my_super_script = MySuperScript(ert)
    my_super_script.run(param_B="param_B", param_A="param_A")

    expected_outputfile = os.path.join(
        config_path,
        "reports",
        user_config_file,
        MySuperScript.__name__,
        SEMEIOSCRIPT_LOG_FILE,
    )

    with open(expected_outputfile) as f:
        published_msgs = f.readlines()
    assert published_msgs[0] == "param_A\n"
    assert published_msgs[1] == "param_B\n"
